"""
Contains the RefGeneSet class

Contains the TCRGene class
"""

from .paths import path_to_db
from collections import OrderedDict
from . import translation
import pandas as pd
import os.path as op

class RefGeneSet:
    """
    RefGeneSet is a class for holding a set of reference TCR sequences

    Attributes
    ----------
    db_file : string
        specifies the name of the file in path_db with reference TCRS
        (e.g alphabeta_db.tsv)
    all_genes : OrderedDict
        contains library of all reference TCRs represented as a dictionary
        of TCRGene instances

    Methods
    -------
    generate_all_genes()
        generates the library of all reference TCRs represented as a dictionary
        of TCRGene instances

    Notes
    -----

    all_genes = RefGeneSet(db_file = "alphabeta_db.tsv").all_genes

    would be equivalent to previously used:

    from all_genes import all_genes


    Information is now callable as follows:
    attribute = 'cdrs'
    RefGeneSet.all_genes['mouse']['TRAV1*01'].__dict__[attribute]
    """
    def __init__(self, db_file):
        self.db_file = db_file
        self.all_genes = self.generate_all_genes(self.db_file)

    def generate_all_genes(self, db_file = "alphabeta_db.tsv"):
        """
        create the reference dataset that will be used throughout tcrdist
        see the TCRGene class below for details

        Parameters
        ----------
        db_file : string
            "alphabeta_db.tsv"

        """
        db_file = op.join(path_to_db, self.db_file)
        assert op.exists(db_file)

        all_genes = OrderedDict()

        db_df = pd.read_csv(db_file, delimiter='\t')

        for rowi, row in db_df.iterrows():
            g = TCRGene( row )
            if g.organism not in all_genes:
                all_genes[g.organism] = OrderedDict() # map from id to TCR_Gene objects
            all_genes[g.organism][g.id] = g
        return(all_genes)


class TCRGene:
    """
    TCRGene is a class for holding information about a single TCR sequence representative

    Parameters
    ----------
    l : list


    Attributes
    ----------
    id : string

    organism : string
        human or mouse
    chain : string
        A or B !!!! this is not the same in TCRrep calls
    region : string
        V or J !!!! this is not the same in TCRrep calls
    protseq: string
        raw protein sequence
    nucseq : string

    alseq : string
        aligned protein sequence
    frame : int

    nucseq_offset : int

    cdr_columns_str : strings
        '28-39;57-66;82-88;106-111',
    cdr_columns : list
        [[28, 39], [57, 66], [82, 88], [106, 111]],
    cdrs : list
        list of cdrs present based on string in the input file l['cdrs']
    cdrs_extracted : list
        list of cdrs extracted based on cdr_column index with gaps retained
    cdrs_aligned : list
        list of cdrs with gaps retained
    cdrs_no_gaps: list
        list of cdrs with gaps removed


    Methods
    -------
    extract_aligned_cdrs()
        wraps _extract_aligned_cdrs
    extract_cdrs_without_gaps()
        wraps _extract_aligned_cdrs

    """
    cdrs_sep = ';'
    gap_character = '.'

    def __init__( self, l):
        self.id       = l['id']
        self.organism = l['organism']
        self.chain    = l['chain']
        self.region   = l['region']
        self.nucseq   = l['nucseq']
        self.alseq    = l['aligned_protseq']
        self.cdr_columns_str = l['cdr_columns']
        self.frame    = l['frame']
        self.rep      = l['id'] # THIS IS PROBABLY NOT ACCUTATE SEE LINE 170 of all genes
        self.mm1_rep  = l['id'] # THIS IS PROBABLY NOT ACCURATE SEE LINES 171 of all genes
        self.nucseq_offset = self.frame - 1 ## 0, 1 or 2 (0-indexed for python)
        self.protseq = translation.get_translation( self.nucseq, self.frame )[0]

        if pd.isnull(l['cdrs']):
            self.cdrs = []
            self.cdr_columns = []
        else:
            self.cdrs = l['cdrs'].split(self.cdrs_sep)
            ## these are still 1-indexed !!!!!!!!!!!!!!
            self.cdr_columns = [ list(map( int, x.split('-'))) for x in l['cdr_columns'].split(self.cdrs_sep) ]


            self.cdrs_extracted = [ self.alseq[ x[0]-1 : x[1] ] for x in self.cdr_columns ]
            self.cdrs_aligned   = self.extract_aligned_cdrs()
            self.cdrs_no_gaps   = self.extract_cdrs_without_gaps()

        # Defensive Assert Statements
        assert self.frame in [1, 2, 3]
        if self.cdrs:
            assert self.cdrs == self.cdrs_extracted # confirm that cdrs in the input file match expectation
            assert self.cdrs_aligned == self.cdrs_extracted

    def extract_aligned_cdrs(self):
        return _extract_aligned_cdrs(self.alseq, self.cdr_columns_str)

    def extract_cdrs_without_gaps(self):
        return _extract_cdrs_without_gaps(self.alseq, self.cdr_columns_str)


def _extract_aligned_cdrs(aligned_protseqs, cdr_columns):
    """

    from db use the aligned_protseq and cdr_column to return the aligned cdrs with gaps preserved


    Parameters
    ----------
    aligned_protseqs : string
        example: 'GQGVEQ.P.AKLMSVEGTFARVNCTYSTSG......FNGLSWYQQREGQAPVFLSYVVL....DGLKDS.....GHFSTFLSRSN.GYSYLLLTELQIKDSASYLCAVR..'
    cdr_columns : string
        example: '28-39;57-66;82-88;106-111')

    Returns
    -------
    cdrs : list


    Examples
    --------
    >>> _extract_aligned_cdrs('GQGVEQ.P.DNLMSVEGTFARVNCTYSTSG......FNGLSWYQQREGHAPVFLSYVVL....DGLKDS.....GHFSTFLSRSN.GYSYLLLTELQIKDSASYLCAVR..',
 '28-39;57-66;82-88;106-111')
    ['TSG......FNG', 'VVL....DGL', 'SRSN.GY', 'CAVR..']
    """
    pos = cdr_columns.split(";")
    pos2 = [list(map(int,x.split("-"))) for x in pos]
    cdrs = [aligned_protseqs[x[0]-1:x[1]] for x in pos2]
    return(cdrs)

def _extract_cdrs_without_gaps(aligned_protseqs, cdr_columns):
    """

    from db use the aligned_protseq and cdr_column to return the aligned cdrs with gaps removed


    Parameters
    ----------
    aligned_protseqs : string
        example: 'GQGVEQ.P.AKLMSVEGTFARVNCTYSTSG......FNGLSWYQQREGQAPVFLSYVVL....DGLKDS.....GHFSTFLSRSN.GYSYLLLTELQIKDSASYLCAVR..'
    cdr_columns : string
        example: '28-39;57-66;82-88;106-111')

    Returns
    -------
    cdrs : list


    Examples
    --------
    >>> _extract_cdrs_without_gaps('GQGVEQ.P.AKLMSVEGTFARVNCTYSTSG......FNGLSWYQQREGQAPVFLSYVVL....DGLKDS.....GHFSTFLSRSN.GYSYLLLTELQIKDSASYLCAVR..',
 '28-39;57-66;82-88;106-111')
    ['TSGFNG', 'VVLDGL', 'SRSNGY', 'CAVR']
    """
    pos = cdr_columns.split(";")
    pos2 = [list(map(int,x.split("-"))) for x in pos]
    cdrs = [aligned_protseqs[x[0]-1:x[1]].replace(".","") for x in pos2]
    return(cdrs)





import pandas as pd
from tcrdist.adpt_funcs import _valid_cdr3
from tcrdist.paths import path_to_base
from tcrdist.repertoire_db import RefGeneSet
import warnings
import os

def import_vdjtools(   vdj_tools_file,
                        chain = 'beta',
                        organism = 'human',
                        db_file = 'alphabeta_gammadelta_db.tsv',
                        validate = True):
    """
    Import VDJtools formated input .tsv or .tsv.gz file. 
    Input file must have columns ['count', 'freq', 'cdr3aa', 'v', 'j','cdr3nt'], 
    see (https://vdjtools-doc.readthedocs.io/en/master/input.html#vdjtools-format)
    for examples of how you can use VDJtool to directly import 
    and convert from the following: 
    MiTCR, MiGEC, IgBlast (MIGMAP), ImmunoSEQ, VDJdb, Vidjil, MiXCR

    Parameters
    -----------
    vdj_tools_file : str
        e.g., os.path.join(path_to_base, 'tcrdist','data','formats','vdj.M_15_CD8_beta.clonotypes.TRB.txt.gz')
    chain : str
        'alpha','beta','gamma', or 'delta'
    organism : str
        'human', 'mouse'
    db_file : str
        'alphabeta_gammadelta_db.tsv',
    validate : bool 
        If True, only clones with valid CDR3AA, V-Gene and J-Gene are returned)
        CDR3 length must be >= 5 for tcrdist3 
    include_nucseq : True
        If True, retain nucletoide sequence

    Returns
    -------
    df or valid_df : DataFrame
        with columns : ['count', 'freq', 'cdr3_b_aa', 'v_b_gene', 'j_b_gene', 'cdr3_b_nucseq','valid_v', 'valid_j', 'valid_cdr3']
    
    """

    assert chain in ['alpha','beta','gamma','delta']
    assert organism in ['human','mouse']
    
    _chain  = { 'alpha' : {'cdr3aa':'cdr3_a_aa','v':'v_a_gene', 'j':'j_a_gene', 'cdr3nt':'cdr3_a_nucseq', 'AB' : 'A'},
                'beta' :  {'cdr3aa':'cdr3_b_aa','v':'v_b_gene', 'j':'j_b_gene', 'cdr3nt':'cdr3_b_nucseq', 'AB' : 'B'},
                'gamma' : {'cdr3aa':'cdr3_g_aa','v':'v_g_gene', 'j':'j_g_gene', 'cdr3nt':'cdr3_g_nucseq', 'AB' : 'A'},
                'delta' : {'cdr3aa':'cdr3_d_aa','v':'v_d_gene', 'j':'j_d_gene', 'cdr3nt':'cdr3_d_nucseq', 'AB' : 'B'} }.\
                get(chain)
    
    
    df = pd.read_csv(vdj_tools_file, sep = "\t")
    
    df = df[['count', 'freq', 'cdr3aa', 'v', 'j','cdr3nt']].\
        rename(columns = {
            'cdr3aa': _chain['cdr3aa'],
            'v'     : _chain['v'],
            'j'     : _chain['j'],
            'cdr3nt': _chain['cdr3nt']
            })

    df[_chain['v']] = df[_chain['v']].apply(lambda x: f"{x}*01")
    df[_chain['j']] = df[_chain['j']].apply(lambda x: f"{x}*01")
        # CHECK AND ADD ALELE
    all_genes =  RefGeneSet(db_file = db_file).all_genes[organism]
    all_valid_genes = [x for x in all_genes.keys() if all_genes[x].chain == _chain['AB']]

    warnings.warn(f"ADDING *01 allele to each V- and J-gene") 
    df['valid_v'] = df[_chain['v']].apply(lambda x : all_genes.get(x) is not None )
    df['valid_j'] = df[_chain['j']].apply(lambda x : all_genes.get(x) is not None )
    df['valid_cdr3'] = df[_chain['cdr3aa']].apply(lambda x : _valid_cdr3(x) and len(x) >=5 )

    x1 = df.shape[0]
    invalid_df = df[~(df['valid_v'] & df['valid_j'] & df['valid_cdr3'])].copy()
    valid_df   = df[(df['valid_v'] & df['valid_j'] & df['valid_cdr3'])].reset_index(drop = True).copy()
    xiv = invalid_df.shape[0]
    warnings.warn(f"{xiv} of {x1} had invalid {_chain['v']}, {_chain['j']}, or {_chain['cdr3aa']}")
    
    if validate:
        warnings.warn(f"REMOVED {xiv} of {x1} with invalid {_chain['v']}, {_chain['j']}, or {_chain['cdr3aa']}")
        return valid_df
    else:
        warnings.warn(f"Invalid clones were not removed, do so manually before proceeding")
        return df
        








import pytest 



def test_workflow_1():
    """
    Load all the TCRs associated with a particular epitope in 
    the Adaptive Biotechnology COVID19 Data Release 2
    """
    import os
    import pandas as pd
    from tcrdist.repertoire import TCRrep

    path = os.path.join('tcrdist', 'data', 'covid19')
    file = 'mira_epitope_9_2477_FLQSINFVR_FLQSINFVRI_FLYLYALVYF_GLEAPFLYLY_INFVRIIMR_LQSINFVRI_LQSINFVRII_QSINFVRII_SINFVRIIMR_VYFLQSINF_VYFLQSINFV_YFLQSINFVR_YLYALVYFL.tcrdist3.csv'
    filename = os.path.join(path,file)
    
    df = pd.read_csv(filename, sep = ",")
   
    df = df[['cell_type','subject','v_b_gene','j_b_gene','cdr3_b_aa',
            'epitope', 'age', 'sex', 'race','cohort',
            'hla-a', 'hla-a_1', 'hla-b', 'hla-b_1', 'hla-c','hla-c_1', 
            'dpa1_1', 'dpb1', 'dpb1_1', 
            'dqa1', 'dqa1_1', 'dqb1','dqb1_1', 'drb1', 'drb1_1', 'drb3']]
    
    df['count'] = 1
    
    tr = TCRrep(cell_df = df, 
                organism = 'human', 
                chains = ['beta'])



"""
example 18
"""

def test_example_report_public_adjustin_attributes ():
    import os
    import pandas as pd
    import numpy as np
    from tcrdist.repertoire import TCRrep
    from tcrdist.public import TCRpublic
    
    # <aesr_fn> antigen-enriched subrepertoire
    aesr_fn = os.path.join(
        'tcrdist',
        'data',
        'covid19',
        'mira_epitope_55_524_ALRKVPTDNYITTY_KVPTDNYITTY.tcrdist3.csv')
    
    # <aesr_df> antigen-enriched subrepertoire
    aesr_df = pd.read_csv(aesr_fn)
    # <tr> TCR repertoire
    tr = TCRrep(
        cell_df = aesr_df[[
            'cohort',
            'subject',
            'v_b_gene', 
            'j_b_gene',
            'cdr3_b_aa']].copy(), 
        organism = 'human', 
        chains = ['beta'], 
        db_file = 'alphabeta_gammadelta_db.tsv', 
        compute_distances = True)
    # <tp> TCRpublic class for reporting publicities 
    tp = TCRpublic(
        tcrrep = tr, 
        output_html_name = "quasi_public_clones2.html")
    # set to True, if we want a universal radius
    tp.fixed_radius = True
    # must then specify maximum distance for finding similar TCRs
    tp.radius = 18
    # set criteria for being quasi-public
    tp.query_str = 'nsubject > 6'
    # Add additional columns to be summarized in the report
    tp.kargs_member_summ['addl_cols'] = ['subject', 'cohort']
    # Add cohort.summary to the labels column so it shows up in the report
    tp.labels.append("cohort.summary")
    # by calling, .report() an html report is made
    public = tp.report()



from setuptools import setup, find_packages
PACKAGES = find_packages()

# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

opts = dict(name='tcrdist3',
            maintainer='Koshlan Mayer-Blackwell',
            maintainer_email='kmayerbl@fredhutch.org',
            description='flexible distance measures for comparing T cell receptors',
            long_description=long_description,
            long_description_content_type='text/markdown',
            url='https://github.com/kmayerb/tcrdist3',
            license='MIT',
            author='Koshlan Mayer-Blackwell',
            author_email='kmayerbl@fredhutch.org',
            version='0.3',
            packages=PACKAGES,
            package_data={"": ["*.csv","*.tsv","*.txt"]},
           )

install_reqs = [
      'pandas>=0.24.2',
      'numpy>=1.16.4',
      'parasail>=1.1.17',
      'scipy>=1.4.1',
      'pwseqdist>=0.6',
      'numba',
      'zipdist>=0.1.5',
      'fishersapi',
      'hierdiff>=0.4',
      'palmotif>=0.2',
      'tcrsampler>=0.1.9',
      'parmap>=1.5.2',
      'olga>=1.2.1',
      'dill>=0.3.2']

if __name__ == "__main__":
      setup(**opts, install_requires=install_reqs)







def test_example_tree_args():
    import os
    import pandas as pd
    from tcrdist.repertoire import TCRrep
    from tcrdist.tree import TCRtree

    df = pd.read_csv("dash.csv").sample(100, random_state=1).reset_index(drop = True)

    tr = TCRrep(cell_df = df, 
                            organism = 'mouse', 
                            chains = ['beta'], 
                            db_file = 'alphabeta_gammadelta_db.tsv')

    tcrtree = TCRtree(tcrrep = tr, 
                html_name = 'dash.mouse.b.tree.html')

    tcrtree.default_hcluster_diff_kwargs = \
            {'clone_df': None,
             'pwmat': None,
             'x_cols': ['epitope'],
             'Z': None,
             'count_col': 'count',
             'subset_ind': None,
             'hclust_method': 'complete',
             'optimal_ordering': True,
             'test_method': 'fishers'}

    tcrtree.default_member_summ_kwargs = \
            {'key_col': 'neighbors_i',
            'count_col': 'count',
            'addl_cols': ['subject'],
            'addl_n': 1}

    tcrtree.default_plot_hclust_props = \
            {'title': '',
            'alpha_col': 'pvalue',
            'alpha': 0.05,
            'tooltip_cols': ['subject',
            'mean_dist',
            'pct_dist_75',
            'pct_dist_50',
            'pct_dist_25',
            'fuzzy_simpson_diversity_75',
            'fuzzy_simpson_diversity_50',
            'fuzzy_simpson_diversity_25',
            'cdr3_b_aa',
            'v_b_gene',
            'j_b_gene',
            'svg_beta',
            'svg_raw_beta',
            'ref_size_beta',
            'ref_unique_beta',
            'percent_missing_beta']}

    tcrtree.build_tree()




import pytest



def test_example_3():
    """
    If want a 'tcrdistances' AND you want control over EVERY parameter.
    """
    import pwseqdist as pw
    import pandas as pd
    from tcrdist.repertoire import TCRrep
    
    df = pd.read_csv("dash.csv")
    tr = TCRrep(cell_df = df, 
                organism = 'mouse', 
                chains = ['alpha','beta'], 
                compute_distances = False,
                db_file = 'alphabeta_gammadelta_db.tsv')

    metrics_a = {
        "cdr3_a_aa" : pw.metrics.nb_vector_tcrdist,
        "pmhc_a_aa" : pw.metrics.nb_vector_tcrdist,
        "cdr2_a_aa" : pw.metrics.nb_vector_tcrdist,
        "cdr1_a_aa" : pw.metrics.nb_vector_tcrdist}

    metrics_b = {
        "cdr3_b_aa" : pw.metrics.nb_vector_tcrdist,
        "pmhc_b_aa" : pw.metrics.nb_vector_tcrdist,
        "cdr2_b_aa" : pw.metrics.nb_vector_tcrdist,
        "cdr1_b_aa" : pw.metrics.nb_vector_tcrdist }

    weights_a= { 
        "cdr3_a_aa" : 3,
        "pmhc_a_aa" : 1,
        "cdr2_a_aa" : 1,
        "cdr1_a_aa" : 1}

    weights_b = { 
        "cdr3_b_aa" : 3,
        "pmhc_b_aa" : 1,
        "cdr2_b_aa" : 1,
        "cdr1_b_aa" : 1}

    kargs_a = {  
        'cdr3_a_aa' : 
            {'use_numba': True, 
            'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 
            'dist_weight': 1, 
            'gap_penalty':4, 
            'ntrim':3, 
            'ctrim':2, 
            'fixed_gappos': False},
        'pmhc_a_aa' : {
            'use_numba': True,
            'distance_matrix': pw.matrices.tcr_nb_distance_matrix,
            'dist_weight':1,
            'gap_penalty':4,
            'ntrim':0,
            'ctrim':0,
            'fixed_gappos':True},
        'cdr2_a_aa' : {
            'use_numba': True,
            'distance_matrix': pw.matrices.tcr_nb_distance_matrix,
            'dist_weight': 1,
            'gap_penalty':4,
            'ntrim':0,
            'ctrim':0,
            'fixed_gappos':True},
        'cdr1_a_aa' : {
            'use_numba': True,
            'distance_matrix': pw.matrices.tcr_nb_distance_matrix,
            'dist_weight':1,
            'gap_penalty':4,
            'ntrim':0,
            'ctrim':0,
            'fixed_gappos':True}
        }
    kargs_b= {  
        'cdr3_b_aa' : 
            {'use_numba': True, 
            'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 
            'dist_weight': 1, 
            'gap_penalty':4, 
            'ntrim':3, 
            'ctrim':2, 
            'fixed_gappos': False},
        'pmhc_b_aa' : {
            'use_numba': True,
            'distance_matrix': pw.matrices.tcr_nb_distance_matrix,
            'dist_weight': 1,
            'gap_penalty':4,
            'ntrim':0,
            'ctrim':0,
            'fixed_gappos':True},
        'cdr2_b_aa' : {
            'use_numba': True,
            'distance_matrix': pw.matrices.tcr_nb_distance_matrix,
            'dist_weight':1,
            'gap_penalty':4,
            'ntrim':0,
            'ctrim':0,
            'fixed_gappos':True},
        'cdr1_b_aa' : {
            'use_numba': True,
            'distance_matrix': pw.matrices.tcr_nb_distance_matrix,
            'dist_weight':1,
            'gap_penalty':4,
            'ntrim':0,
            'ctrim':0,
            'fixed_gappos':True}
        }   

    tr.metrics_a = metrics_a
    tr.metrics_b = metrics_b

    tr.weights_a = weights_a
    tr.weights_b = weights_b

    tr.kargs_a = kargs_a 
    tr.kargs_b = kargs_b

    tr.compute_distances()
    tr.pw_alpha
    tr.pw_beta


"""
mappers
=======

The mappers module contain dictionaries and methods for taking common output and
reformatting it to match tcrdist2 input requirements.

The top of the file contains ordered dictionaries mapping keys

More complex mappings are encapsulated in methods to be called on pd.DataFrames.

"""
import ast
import pandas as pd
from collections import OrderedDict
from functools import reduce
import warnings

tcrdist_to_tcrdist2_mapping = OrderedDict([('id', 'id'),
                                           ('epitope', 'epitope'),
                                           ('subject', 'subject'),
                                           ('cdr3a'  , 'cdr3_a_aa'),
                                           ('cdr3b'  , 'cdr3_b_aa'),
                                           ('ja_gene', 'j_a_gene'),
                                           ('va_gene', 'v_a_gene'),
                                           ('jb_gene', 'j_b_gene'),
                                           ('vb_gene', 'v_b_gene')])


vdjdb_to_tcrdist2_mapping_TRA = OrderedDict([('complex.id', 'complex_id'),
                                        ('Gene', 'gene'),
                                        ('CDR3', 'cdr3_a_aa'),
                                        ('V', 'v_a_gene'),
                                        ('J', 'j_a_gene'),
                                        ('Species', 'organism'),
                                        ('MHC A', 'mhc_a_a'),
                                        ('MHC B', 'mhc_a_b'),
                                        ('MHC class', 'mhc_a_class'),
                                        ('Epitope', 'epitope'),
                                        ('Epitope gene', 'epitope_gene'),
                                        ('Epitope species', 'epitope_species'),
                                        ('Reference', 'reference'),
                                        ('Method', 'method'),
                                        ('Meta', 'meta'),
                                        ('CDR3fix', 'cdr3fix'),
                                        ('Score', 'score')])

vdjdb_to_tcrdist2_mapping_TRB = OrderedDict([('complex.id', 'complex_id'),
                                        ('Gene', 'gene'),
                                        ('CDR3', 'cdr3_b_aa'),
                                        ('V', 'v_b_gene'),
                                        ('J', 'j_b_gene'),
                                        ('Species', 'organism'),
                                        ('MHC A', 'mhc_b_a'),
                                        ('MHC B', 'mhc_b_b'),
                                        ('MHC class', 'mhc_b_class'),
                                        ('Epitope', 'epitope'),
                                        ('Epitope gene', 'epitope_gene'),
                                        ('Epitope species', 'epitope_species'),
                                        ('Reference', 'reference'),
                                        ('Method', 'method'),
                                        ('Meta', 'meta'),
                                        ('CDR3fix', 'cdr3fix'),
                                        ('Score', 'score')])


tcrdist_clone_df_to_tcrdist2_mapping = OrderedDict([ ('clone_id'  , 'clone_id' ),
                                                     ('subject'   , 'subject'  ),
                                                     ('cdr3a'     , 'cdr3_a_aa'),
                                                     ('cdr3b'     , 'cdr3_b_aa'),
                                                     ('clone_size', 'count'    ),
                                                     ('epitope'   , 'epitope'  ),
                                                     ('ja_rep'    , 'j_a_gene' ),
                                                     ('jb_rep'    , 'j_b_gene' ),
                                                     ('subject'   , 'subject'  ),
                                                     ('va_rep'    , 'v_a_gene' ),
                                                     ('vb_rep'    , 'v_b_gene' ),
                                                     ('cdr3a_nucseq', 'cdr3_a_nucseq'),
                                                     ('cdr3b_nucseq', 'cdr3_b_nucseq'),
                                                     ('va_countreps', 'va_countreps'),
                                                     ('ja_countreps', 'ja_countreps'),
                                                     ('vb_countreps', 'vb_countreps'),
                                                     ('jb_countreps', 'jb_countreps'),
                                                     ('va_gene'     , 'va_gene'),
                                                     ('ja_gene'     , 'ja_gene'),
                                                     ('vb_gene'     , 'vb_gene'),
                                                     ('jb_gene'     , 'jb_gene')])

tcrdist2_to_tcrdist_clone_df_mapping_gd_red = OrderedDict([('clone_id', 'clone_id'),
                                                        ('subject', 'subject'),
                                                        ('cdr3_g_aa', 'cdr3a'),
                                                        ('cdr3_d_aa', 'cdr3b'),
                                                        ('count', 'clone_size'),
                                                        ('epitope', 'epitope'),
                                                        ('j_g_gene', 'ja_rep'),
                                                        ('j_d_gene', 'jb_rep'),
                                                        ('v_g_gene', 'va_rep'),
                                                        ('v_d_gene', 'vb_rep'),
                                                        ('cdr3_g_nucseq', 'cdr3a_nucseq'),
                                                        ('cdr3_d_nucseq', 'cdr3b_nucseq'),
                                                        ('vg_countreps', 'va_countreps'),
                                                        ('jg_countreps', 'ja_countreps'),
                                                        ('vd_countreps', 'vb_countreps'),
                                                        ('jd_countreps', 'jb_countreps'),
                                                        ('vg_gene'     , 'va_gene'),
                                                        ('jg_gene'     , 'ja_gene'),
                                                        ('vd_gene'     , 'vb_gene'),
                                                        ('jd_gene'     , 'jb_gene'),
                                                        ('va_countreps', 'va_countreps'),
                                                        ('ja_countreps', 'ja_countreps'),
                                                        ('vb_countreps', 'vb_countreps'),
                                                        ('jb_countreps', 'jb_countreps'),
                                                        ('va_gene'     , 'va_gene'),
                                                        ('ja_gene'     , 'ja_gene'),
                                                        ('vb_gene'     , 'vb_gene'),
                                                        ('jb_gene'     , 'jb_gene')])

tcrdist2_to_tcrdist_clone_df_mapping_gd = OrderedDict([('clone_id', 'clone_id'),
                                                        ('subject', 'subject'),
                                                        ('cdr3_g_aa', 'cdr3a'),
                                                        ('cdr3_d_aa', 'cdr3b'),
                                                        ('count', 'clone_size'),
                                                        ('epitope', 'epitope'),
                                                        ('j_g_gene', 'ja_rep'),
                                                        ('j_d_gene', 'jb_rep'),
                                                        ('v_g_gene', 'va_rep'),
                                                        ('v_d_gene', 'vb_rep'),
                                                        ('cdr3_g_nucseq', 'cdr3a_nucseq'),
                                                        ('cdr3_d_nucseq', 'cdr3b_nucseq'),
                                                        ('va_countreps', 'va_countreps'),
                                                        ('ja_countreps', 'ja_countreps'),
                                                        ('vb_countreps', 'vb_countreps'),
                                                        ('jb_countreps', 'jb_countreps'),
                                                        ('va_gene'     , 'va_gene'),
                                                        ('ja_gene'     , 'ja_gene'),
                                                        ('vb_gene'     , 'vb_gene'),
                                                        ('jb_gene'     , 'jb_gene')])



tcrdist2_to_tcrdist_clone_df_mapping = OrderedDict([('clone_id', 'clone_id'),
                                                     ('subject', 'subject'),
                                                     ('cdr3_a_aa', 'cdr3a'),
                                                     ('cdr3_b_aa', 'cdr3b'),
                                                     ('count', 'clone_size'),
                                                     ('epitope', 'epitope'),
                                                     ('j_a_gene', 'ja_rep'),
                                                     ('j_b_gene', 'jb_rep'),
                                                     ('v_a_gene', 'va_rep'),
                                                     ('v_b_gene', 'vb_rep'),
                                                     ('cdr3_a_nucseq', 'cdr3a_nucseq'),
                                                     ('cdr3_b_nucseq', 'cdr3b_nucseq'),
                                                     ('va_countreps', 'va_countreps'),
                                                     ('ja_countreps', 'ja_countreps'),
                                                     ('vb_countreps', 'vb_countreps'),
                                                     ('jb_countreps', 'jb_countreps'),
                                                     ('va_gene'     , 'va_gene'),
                                                     ('ja_gene'     , 'ja_gene'),
                                                     ('vb_gene'     , 'vb_gene'),
                                                     ('jb_gene'     , 'jb_gene')])






tcrdist2_to_tcrdist_clone_df_mapping_a = OrderedDict([('clone_id', 'clone_id'),
                                                     ('subject', 'subject'),
                                                     ('cdr3_a_aa', 'cdr3a'),
                                                     ('count', 'clone_size'),
                                                     ('epitope', 'epitope'),
                                                     ('j_a_gene', 'ja_rep'),
                                                     ('v_a_gene', 'va_rep'),
                                                     ('cdr3_a_nucseq', 'cdr3a_nucseq'),
                                                     ('cdr3_b_nucseq', 'cdr3b_nucseq'),
                                                     ('va_countreps', 'va_countreps'),
                                                     ('ja_countreps', 'ja_countreps'),
                                                     ('va_gene'     , 'va_gene'),
                                                     ('ja_gene'     , 'ja_gene')])


tcrdist2_to_tcrdist_clone_df_mapping_b = OrderedDict([('clone_id', 'clone_id'),
                                                     ('subject', 'subject'),
                                                     ('cdr3_b_aa', 'cdr3b'),
                                                     ('count', 'clone_size'),
                                                     ('epitope', 'epitope'),
                                                     ('j_b_gene', 'jb_rep'),
                                                     ('v_b_gene', 'vb_rep'),
                                                     ('cdr3_b_nucseq', 'cdr3b_nucseq'),
                                                     ('vb_countreps', 'vb_countreps'),
                                                     ('jb_countreps', 'jb_countreps'),
                                                     ('vb_gene'     , 'vb_gene'),
                                                     ('jb_gene'     , 'jb_gene')])



# Note that original tcrdist had no concept of gamma and delta syntax g->a , d->b 


tcrdist2_to_tcrdist_clone_df_mapping_g = OrderedDict([('clone_id', 'clone_id'),
                                                        ('subject', 'subject'),
                                                        ('cdr3_g_aa', 'cdr3a'),
                                                        ('count', 'clone_size'),
                                                        ('epitope', 'epitope'),
                                                        ('j_g_gene', 'ja_rep'),
                                                        ('v_g_gene', 'va_rep'),
                                                        ('cdr3_g_nucseq', 'cdr3a_nucseq'),
                                                        ('va_countreps', 'va_countreps'),
                                                        ('ja_countreps', 'ja_countreps'),
                                                        ('va_gene'     , 'va_gene'),
                                                        ('ja_gene'     , 'ja_gene')])

tcrdist2_to_tcrdist_clone_df_mapping_d = OrderedDict([('clone_id', 'clone_id'),
                                                        ('subject', 'subject'),
                                                        ('cdr3_d_aa', 'cdr3b'),
                                                        ('count', 'clone_size'),
                                                        ('epitope', 'epitope'),
                                                        ('j_d_gene', 'jb_rep'),
                                                        ('v_d_gene', 'vb_rep'),
                                                        ('cdr3_d_nucseq', 'cdr3b_nucseq'),
                                                        ('vb_countreps', 'vb_countreps'),
                                                        ('jb_countreps', 'jb_countreps'),
                                                        ('vb_gene'     , 'vb_gene'),
                                                        ('jb_gene'     , 'jb_gene')])

TCRrep_clone_df_to_TCRMotif_clone_df =OrderedDict([ ('subject'   , 'subject'),
                                                    ('epitope'  , 'epitope'),
                                                    ('v_a_gene' , 'va_rep'),
                                                    ('j_a_gene' , 'ja_rep'),
                                                    ('v_b_gene' , 'vb_rep'),
                                                    ('j_b_gene' , 'jb_rep'),
                                                    ('cdr3_a_aa', 'cdr3a'),
                                                    ('cdr3_b_aa', 'cdr3b' ) ])

TCRsubset_clone_df_to_TCRMotif_clone_df  = OrderedDict([('subject', 'subject'),
                                                        ('epitope', 'epitope'),
                                                        ('v_a_gene', 'va_rep'),
                                                        ('j_a_gene', 'ja_rep'),
                                                        ('v_b_gene', 'vb_rep'),
                                                        ('j_b_gene', 'jb_rep'),
                                                        ('cdr3_a_aa', 'cdr3a'),
                                                        ('cdr3_b_aa', 'cdr3b')])
"""gd_redundand stands for gamma delta redundant mapper"""
TCRsubset_clone_df_to_TCRMotif_clone_df_gd_red  = OrderedDict([('subject', 'subject'),
                                                                ('epitope', 'epitope'),
                                                                ('v_a_gene', 'va_rep'),
                                                                ('v_g_gene', 'va_rep'),
                                                                ('j_a_gene', 'ja_rep'),
                                                                ('j_g_gene', 'ja_rep'),
                                                                ('v_b_gene', 'vb_rep'),
                                                                ('v_d_gene', 'vb_rep'),
                                                                ('j_b_gene', 'jb_rep'),
                                                                ('j_d_gene', 'jb_rep'),
                                                                ('cdr3_a_aa', 'cdr3a'),
                                                                ('cdr3_g_aa', 'cdr3a'),
                                                                ('cdr3_b_aa', 'cdr3b'),
                                                                ('cdr3_d_aa', 'cdr3b')])


def generic_pandas_mapper(df,mapping, allow_missing = False):
    """
    Parameters
    ----------
    df : DataFrame
        Input DataFrame
    mapping : OrderedDict
        defines columns to select and rename

    Returns
    -------
    df2 : DataFrame
        Renamed DataFrame

    Example
    -------
    >>> import pandas as pd
    >>> from collections import OrderedDict
    >>> df = pd.DataFrame({"a":[1,2],"b":[3,4]})
    >>> mapper = OrderedDict([('a', 'ardvark')])
    >>> generic_pandas_mapping(df, mapper)
           ardvark
    0        1
    1        2
    """
    if not isinstance(df, pd.DataFrame):
        raise TypeError("df must be a Pandas DataFrame")
    if not isinstance(mapping, OrderedDict):
        raise TypeError('mapping must be and OrderedDict')
    if not all([elem in df.keys() for elem in list(mapping.keys())]):
        if allow_missing:
            mapping = {k:v for k,v in mapping.items() if k in df.keys()}
        else:
            raise KeyError('all mapping keys must be columns of input df')

    df2 = df[mapping.keys()].rename(columns = mapping).copy()
    return(df2)

def vdjdb_to_tcrdist2(pd_df):
    """
    reformat .tsv downloaded from vdjdb (July 2019) to input format of tcrdist2 input format

    Parameters
    ----------
    pd_df : pandas.core.DataFrame likley derived from pd.read_csv() on .tsv from
    vdjdb.org

    Returns
    -------
    tcrdist2_formatted_pd_df

    Notes
    -----
    Code is written for clarity rather than length or computatinal efficiency
    since parsing demands may change.

    Note that in the unpack phase in the fist block of code a dictionary is made
    to hold all features in the vdjdb output. Much of the information from
    paired sequences is redundant (i.e. epitope). If memory demands increase
    for larger number of rows, consider modifying to unpack less of the output.

    Requires: pandas and ast (abstract syntax trees)

    Raises
    ------
    AssertionError if 'Gene' field is not TRA and TRB

    TODO
    ----
    Handle case of missing alpha or beta seq


    """
    d = {}
    # unpack vdjdb panda.DataFrame to dictionary
    for index, row in pd_df.iterrows():

        # gene will be either TRA or
        gene = row['Gene']
        assert(gene in ["TRA", "TRB"]), "Unexpected Input in vdjdb 'Gene' field not TRA or TRB. Check input."

        # complex_id is shared for paired chains
        complex_id = row['complex.id']

        # note: ast (abstract syntax trees) allows converstion of string to dictionary)
        # packed within a pd.DF cell
        meta_dict =  ast.literal_eval(row['Meta'])
        method_dict = ast.literal_eval(row['Method'])

        if gene == "TRA":
            row = row.rename(vdjdb_to_tcrdist2_mapping_TRA)
        elif gene == "TRB":
            row = row.rename(vdjdb_to_tcrdist2_mapping_TRB)

        d.setdefault(complex_id, {})[gene] = row.to_dict()
        d[complex_id][gene].update(meta_dict)
        d[complex_id][gene].update(method_dict)


    # output select fields to a list of dictionaries (l_out)
    complex_ids = sorted(d.keys())
    l_out = []
    for complex_id in complex_ids:
        try:
            id         = d[complex_id]["TRA"]['complex_id']
            cell_type  = d[complex_id]["TRA"]['cell.subset']
            organism   = d[complex_id]["TRA"]['organism']
            epitope_aa = d[complex_id]["TRA"]['epitope']
            epitope    = d[complex_id]["TRA"]['epitope.id']
            subject    = d[complex_id]["TRA"]['subject.id']

            mhc_a_a    = d[complex_id]["TRA"]['mhc_a_a']
            mhc_a_b    = d[complex_id]["TRA"]['mhc_a_b']

            mhc_b_a    = d[complex_id]["TRB"]['mhc_b_a']
            mhc_b_b    = d[complex_id]["TRB"]['mhc_b_b']

            cdr3_a_aa  = d[complex_id]["TRA"]['cdr3_a_aa']
            v_a_gene   = d[complex_id]["TRA"]['v_a_gene']
            j_a_gene   = d[complex_id]["TRA"]['j_a_gene']

            cdr3_b_aa  = d[complex_id]["TRB"]['cdr3_b_aa']
            v_b_gene   = d[complex_id]["TRB"]['v_b_gene']
            j_b_gene   = d[complex_id]["TRB"]['j_b_gene']

            frequency  = d[complex_id]["TRA"]["frequency"]
            try:
                count      = int(frequency.split("/")[0])
            except ValueError:
                count      = 1
            l_out.append(
            {'id'        : id,
            'cell_type'  : cell_type,
            'organism'   : organism,
            'epitope_aa' : epitope_aa,
            'epitope'    : epitope,
            'subject'    : subject,
            'mhc_a_a'    : mhc_a_a,
            'mhc_a_b'    : mhc_a_b,
            'mhc_b_a'    : mhc_b_a,
            'mhc_b_b'    : mhc_b_b,
            'cdr3_a_aa'  : cdr3_a_aa,
            'v_a_gene'   : v_a_gene,
            'j_a_gene'   : j_a_gene,
            'cdr3_b_aa'  : cdr3_b_aa,
            'v_b_gene'   : v_b_gene,
            'j_b_gene'   : j_b_gene,
            'frequency'  : frequency,
            'count'      : count})
        except KeyError:
            pass


    # convert list of dictionaries to pandas DataFrame
    tcrdist2_formatted_pd_df = pd.DataFrame.from_dict(l_out)
    return(tcrdist2_formatted_pd_df)



def multistudy_vdjdb_to_tcrdist2(pd_df):
    """

    map vdjdb data from multiple studies to one tcrdist2 formatted pd.dataframe

    Parameters
    ----------
    pd_df : Pandas DataFrame
        from vdjdb.org with headers as of July 10, 2019

    Returns
    -------

    tcr_formatted_dfs_merged : Pandas DataFrame
        with all paired data from input, single combined data frame tcrdist2 mapped

    """
    # 2 complex.id which links paired sequences is Reference (study specific)
    studies = pd_df['Reference'].unique()

    # break full df into sub dfs: one per study
    dfs_split_by_study = {study: pd.DataFrame for study in studies}
    for study in list(dfs_split_by_study.keys()):
        dfs_split_by_study[study] = pd_df[:][pd_df.Reference == study].reset_index()

    # tcrdist format each study specific dataframe
    tcr_formatted_dfs_split_by_study = {study: None for study in studies}
    for study in list(dfs_split_by_study.keys()):
        tcr_formatted_dfs_split_by_study[study] = vdjdb_to_tcrdist2(pd_df = dfs_split_by_study[study])

    # create list of studies with paired data
    studies_with_AB_data = [k for k in list(tcr_formatted_dfs_split_by_study.keys()) if \
                            tcr_formatted_dfs_split_by_study[k].shape[0] != 0]
    studies_without_AB_data = [k for k in list(tcr_formatted_dfs_split_by_study.keys()) if \
                               tcr_formatted_dfs_split_by_study[k].shape[0] != 0]

    # subset dictionary of dataframes to those with paired data
    ab = { k: tcr_formatted_dfs_split_by_study[k] for k in studies_with_AB_data }

    # count rows in each dataset for human interest
    ab_rows = [x[1].shape[0] for x in list(ab.items())]

    # convert dictionary of dataframes into a list and only keep the dataframe ignoring the study name.
    def append_study_name(name, df):
        df["reference"] = name
        return(df)

    ab_dfs = [append_study_name(x[0], x[1]) for x in list(ab.items())]

    # combine the list of dataframes into a single dataframe
    tcr_formatted_dfs_merged = reduce(lambda  x, y: pd.concat([x, y]), ab_dfs)

    return tcr_formatted_dfs_merged


def map_genes_to_first_alleles(genes, organism):
    """
    returns a list of the first alleles ('TRAJ11*01') for a given gene (TRAJ11)

    Parameters
    ----------
    genes : lists
        list of gene names using IMGT Nomenclature
    organism : string
        'human' or 'mouse'9

    Returns
    -------
    alleles : list

    Notes
    -----

    The mapping dictionary was produced with the following code
    from collections import OrderedDict
    pd_df = pd.read_csv('tcrdist/db/alphabeta_db.tsv', sep = '\t')
    g = pd_df.id.str.split("*")
    gene = [x[0] for x in g]
    X = pd.DataFrame({'gene' : gene, 'organism' : pd_df.organism.copy(), 'id' : pd_df.id.copy()})
    X = X.groupby(['gene', 'organism']).first().reset_index()
    d1 = {"mouse": OrderedDict(), "human" : OrderedDict()}
    for row in X.iterrows():
        d1[row[1]['organism']][row[1]['gene']] = row[1]['id']

    # Load second file
    pd_df = pd.read_csv('tcrdist/db/gammadelta_db.tsv', sep = '\t')
    g = pd_df.id.str.split("*")
    gene = [x[0] for x in g]
    X = pd.DataFrame({'gene' : gene, 'organism' : pd_df.organism.copy(), 'id' : pd_df.id.copy()})
    X = X.groupby(['gene', 'organism']).first().reset_index()
    for row in X.iterrows():
        d1[row[1]['organism']][row[1]['gene']] = row[1]['id']

    """

    d = {'mouse': OrderedDict([('TRAJ11', 'TRAJ11*01'), ('TRAJ12', 'TRAJ12*01'), ('TRAJ13', 'TRAJ13*01'), ('TRAJ15', 'TRAJ15*01'), ('TRAJ16', 'TRAJ16*01'), ('TRAJ17', 'TRAJ17*01'), ('TRAJ18', 'TRAJ18*01'), ('TRAJ19', 'TRAJ19*01'), ('TRAJ2', 'TRAJ2*01'), ('TRAJ20', 'TRAJ20*01'), ('TRAJ21', 'TRAJ21*01'), ('TRAJ22', 'TRAJ22*01'), ('TRAJ23', 'TRAJ23*01'), ('TRAJ24', 'TRAJ24*01'), ('TRAJ25', 'TRAJ25*01'), ('TRAJ26', 'TRAJ26*01'), ('TRAJ27', 'TRAJ27*01'), ('TRAJ28', 'TRAJ28*01'), ('TRAJ29', 'TRAJ29*01'), ('TRAJ3', 'TRAJ3*01'), ('TRAJ30', 'TRAJ30*01'), ('TRAJ31', 'TRAJ31*01'), ('TRAJ32', 'TRAJ32*01'), ('TRAJ33', 'TRAJ33*01'), ('TRAJ34', 'TRAJ34*01'), ('TRAJ35', 'TRAJ35*01'), ('TRAJ36', 'TRAJ36*01'), ('TRAJ37', 'TRAJ37*01'), ('TRAJ38', 'TRAJ38*01'), ('TRAJ39', 'TRAJ39*01'), ('TRAJ4', 'TRAJ4*01'), ('TRAJ40', 'TRAJ40*01'), ('TRAJ41', 'TRAJ41*01'), ('TRAJ42', 'TRAJ42*01'), ('TRAJ43', 'TRAJ43*01'), ('TRAJ44', 'TRAJ44*01'), ('TRAJ45', 'TRAJ45*01'), ('TRAJ46', 'TRAJ46*01'), ('TRAJ47', 'TRAJ47*01'), ('TRAJ48', 'TRAJ48*01'), ('TRAJ49', 'TRAJ49*01'), ('TRAJ5', 'TRAJ5*01'), ('TRAJ50', 'TRAJ50*01'), ('TRAJ52', 'TRAJ52*01'), ('TRAJ53', 'TRAJ53*01'), ('TRAJ54', 'TRAJ54*01'), ('TRAJ56', 'TRAJ56*01'), ('TRAJ57', 'TRAJ57*01'), ('TRAJ58', 'TRAJ58*01'), ('TRAJ59', 'TRAJ59*01'), ('TRAJ6', 'TRAJ6*01'), ('TRAJ60', 'TRAJ60*01'), ('TRAJ61', 'TRAJ61*01'), ('TRAJ7', 'TRAJ7*01'), ('TRAJ9', 'TRAJ9*01'), ('TRAV1', 'TRAV1*01'), ('TRAV10', 'TRAV10*01'), ('TRAV10D', 'TRAV10D*01'), ('TRAV10N', 'TRAV10N*01'), ('TRAV11', 'TRAV11*01'), ('TRAV11D', 'TRAV11D*01'), ('TRAV11N', 'TRAV11N*01'), ('TRAV12-1', 'TRAV12-1*01'), ('TRAV12-2', 'TRAV12-2*01'), ('TRAV12-3', 'TRAV12-3*01'), ('TRAV12D-1', 'TRAV12D-1*01'), ('TRAV12D-2', 'TRAV12D-2*01'), ('TRAV12D-3', 'TRAV12D-3*01'), ('TRAV12N-1', 'TRAV12N-1*01'), ('TRAV12N-2', 'TRAV12N-2*01'), ('TRAV12N-3', 'TRAV12N-3*01'), ('TRAV13-1', 'TRAV13-1*01'), ('TRAV13-2', 'TRAV13-2*01'), ('TRAV13-3', 'TRAV13-3*01'), ('TRAV13-4/DV7', 'TRAV13-4/DV7*01'), ('TRAV13-5', 'TRAV13-5*01'), ('TRAV13D-1', 'TRAV13D-1*01'), ('TRAV13D-2', 'TRAV13D-2*01'), ('TRAV13D-3', 'TRAV13D-3*01'), ('TRAV13D-4', 'TRAV13D-4*01'), ('TRAV13N-1', 'TRAV13N-1*01'), ('TRAV13N-2', 'TRAV13N-2*01'), ('TRAV13N-3', 'TRAV13N-3*01'), ('TRAV13N-4', 'TRAV13N-4*01'), ('TRAV14-1', 'TRAV14-1*01'), ('TRAV14-2', 'TRAV14-2*01'), ('TRAV14-3', 'TRAV14-3*01'), ('TRAV14D-1', 'TRAV14D-1*01'), ('TRAV14D-2', 'TRAV14D-2*01'), ('TRAV14D-3/DV8', 'TRAV14D-3/DV8*01'), ('TRAV14N-1', 'TRAV14N-1*01'), ('TRAV14N-2', 'TRAV14N-2*01'), ('TRAV14N-3', 'TRAV14N-3*01'), ('TRAV15-1/DV6-1', 'TRAV15-1/DV6-1*01'), ('TRAV15-2/DV6-2', 'TRAV15-2/DV6-2*01'), ('TRAV15D-1/DV6D-1', 'TRAV15D-1/DV6D-1*01'), ('TRAV15D-2/DV6D-2', 'TRAV15D-2/DV6D-2*01'), ('TRAV15N-1', 'TRAV15N-1*01'), ('TRAV15N-2', 'TRAV15N-2*01'), ('TRAV16', 'TRAV16*01'), ('TRAV16D/DV11', 'TRAV16D/DV11*01'), ('TRAV16N', 'TRAV16N*01'), ('TRAV17', 'TRAV17*01'), ('TRAV18', 'TRAV18*01'), ('TRAV19', 'TRAV19*01'), ('TRAV2', 'TRAV2*01'), ('TRAV20', 'TRAV20*01'), ('TRAV21/DV12', 'TRAV21/DV12*01'), ('TRAV3-1', 'TRAV3-1*01'), ('TRAV3-3', 'TRAV3-3*01'), ('TRAV3-4', 'TRAV3-4*01'), ('TRAV3D-3', 'TRAV3D-3*01'), ('TRAV3N-3', 'TRAV3N-3*01'), ('TRAV4-2', 'TRAV4-2*01'), ('TRAV4-3', 'TRAV4-3*01'), ('TRAV4-4/DV10', 'TRAV4-4/DV10*01'), ('TRAV4D-2', 'TRAV4D-2*01'), ('TRAV4D-3', 'TRAV4D-3*01'), ('TRAV4D-4', 'TRAV4D-4*01'), ('TRAV4N-3', 'TRAV4N-3*01'), ('TRAV4N-4', 'TRAV4N-4*01'), ('TRAV5-1', 'TRAV5-1*01'), ('TRAV5-2', 'TRAV5-2*01'), ('TRAV5-4', 'TRAV5-4*01'), ('TRAV5D-2', 'TRAV5D-2*01'), ('TRAV5D-4', 'TRAV5D-4*01'), ('TRAV5N-2', 'TRAV5N-2*01'), ('TRAV5N-4', 'TRAV5N-4*01'), ('TRAV6-1', 'TRAV6-1*01'), ('TRAV6-2', 'TRAV6-2*01'), ('TRAV6-3', 'TRAV6-3*01'), ('TRAV6-4', 'TRAV6-4*01'), ('TRAV6-5', 'TRAV6-5*01'), ('TRAV6-6', 'TRAV6-6*01'), ('TRAV6-7/DV9', 'TRAV6-7/DV9*01'), ('TRAV6D-3', 'TRAV6D-3*01'), ('TRAV6D-4', 'TRAV6D-4*01'), ('TRAV6D-5', 'TRAV6D-5*01'), ('TRAV6D-6', 'TRAV6D-6*01'), ('TRAV6D-7', 'TRAV6D-7*01'), ('TRAV6N-5', 'TRAV6N-5*01'), ('TRAV6N-6', 'TRAV6N-6*01'), ('TRAV6N-7', 'TRAV6N-7*01'), ('TRAV7-1', 'TRAV7-1*01'), ('TRAV7-2', 'TRAV7-2*01'), ('TRAV7-3', 'TRAV7-3*01'), ('TRAV7-4', 'TRAV7-4*01'), ('TRAV7-5', 'TRAV7-5*01'), ('TRAV7-6', 'TRAV7-6*01'), ('TRAV7D-2', 'TRAV7D-2*01'), ('TRAV7D-3', 'TRAV7D-3*01'), ('TRAV7D-4', 'TRAV7D-4*01'), ('TRAV7D-5', 'TRAV7D-5*01'), ('TRAV7D-6', 'TRAV7D-6*01'), ('TRAV7N-4', 'TRAV7N-4*01'), ('TRAV7N-5', 'TRAV7N-5*01'), ('TRAV7N-6', 'TRAV7N-6*01'), ('TRAV8-1', 'TRAV8-1*01'), ('TRAV8-2', 'TRAV8-2*01'), ('TRAV8D-1', 'TRAV8D-1*01'), ('TRAV8D-2', 'TRAV8D-2*01'), ('TRAV8N-2', 'TRAV8N-2*01'), ('TRAV9-1', 'TRAV9-1*01'), ('TRAV9-2', 'TRAV9-2*01'), ('TRAV9-3', 'TRAV9-3*01'), ('TRAV9-4', 'TRAV9-4*01'), ('TRAV9D-1', 'TRAV9D-1*01'), ('TRAV9D-2', 'TRAV9D-2*01'), ('TRAV9D-3', 'TRAV9D-3*01'), ('TRAV9D-4', 'TRAV9D-4*01'), ('TRAV9N-2', 'TRAV9N-2*01'), ('TRAV9N-3', 'TRAV9N-3*01'), ('TRAV9N-4', 'TRAV9N-4*01'), ('TRBD1', 'TRBD1*01'), ('TRBD2', 'TRBD2*01'), ('TRBJ1-1', 'TRBJ1-1*01'), ('TRBJ1-2', 'TRBJ1-2*01'), ('TRBJ1-3', 'TRBJ1-3*01'), ('TRBJ1-4', 'TRBJ1-4*01'), ('TRBJ1-5', 'TRBJ1-5*01'), ('TRBJ1-6', 'TRBJ1-6*01'), ('TRBJ1-7', 'TRBJ1-7*01'), ('TRBJ2-1', 'TRBJ2-1*01'), ('TRBJ2-2', 'TRBJ2-2*01'), ('TRBJ2-3', 'TRBJ2-3*01'), ('TRBJ2-4', 'TRBJ2-4*01'), ('TRBJ2-5', 'TRBJ2-5*01'), ('TRBJ2-6', 'TRBJ2-6*01'), ('TRBJ2-7', 'TRBJ2-7*01'), ('TRBV1', 'TRBV1*01'), ('TRBV10', 'TRBV10*01'), ('TRBV12-1', 'TRBV12-1*01'), ('TRBV12-2', 'TRBV12-2*01'), ('TRBV13-1', 'TRBV13-1*01'), ('TRBV13-2', 'TRBV13-2*01'), ('TRBV13-3', 'TRBV13-3*01'), ('TRBV14', 'TRBV14*01'), ('TRBV15', 'TRBV15*01'), ('TRBV16', 'TRBV16*01'), ('TRBV17', 'TRBV17*01'), ('TRBV19', 'TRBV19*01'), ('TRBV2', 'TRBV2*01'), ('TRBV20', 'TRBV20*01'), ('TRBV21', 'TRBV21*01'), ('TRBV23', 'TRBV23*01'), ('TRBV24', 'TRBV24*01'), ('TRBV26', 'TRBV26*01'), ('TRBV29', 'TRBV29*01'), ('TRBV3', 'TRBV3*01'), ('TRBV30', 'TRBV30*01'), ('TRBV31', 'TRBV31*01'), ('TRBV4', 'TRBV4*01'), ('TRBV5', 'TRBV5*01'), ('TRBV8', 'TRBV8*01'), ('TRBV9', 'TRBV9*01'), ('TRDD1', 'TRDD1*01'), ('TRDD2', 'TRDD2*01'), ('TRDJ1', 'TRDJ1*01'), ('TRDJ2', 'TRDJ2*01'), ('TRDV1', 'TRDV1*01'), ('TRDV2-1', 'TRDV2-1*01'), ('TRDV2-2', 'TRDV2-2*01'), ('TRDV4', 'TRDV4*01'), ('TRDV5', 'TRDV5*01'), ('TRGJ1', 'TRGJ1*01'), ('TRGJ2', 'TRGJ2*01'), ('TRGJ3', 'TRGJ3*01'), ('TRGJ4', 'TRGJ4*01'), ('TRGV1', 'TRGV1*01'), ('TRGV2', 'TRGV2*01'), ('TRGV3', 'TRGV3*01'), ('TRGV4', 'TRGV4*01'), ('TRGV5', 'TRGV5*01'), ('TRGV6', 'TRGV6*01'), ('TRGV7', 'TRGV7*01')]), 'human': OrderedDict([('TRAJ1', 'TRAJ1*01'), ('TRAJ10', 'TRAJ10*01'), ('TRAJ11', 'TRAJ11*01'), ('TRAJ12', 'TRAJ12*01'), ('TRAJ13', 'TRAJ13*01'), ('TRAJ14', 'TRAJ14*01'), ('TRAJ15', 'TRAJ15*01'), ('TRAJ16', 'TRAJ16*01'), ('TRAJ17', 'TRAJ17*01'), ('TRAJ18', 'TRAJ18*01'), ('TRAJ19', 'TRAJ19*01'), ('TRAJ2', 'TRAJ2*01'), ('TRAJ20', 'TRAJ20*01'), ('TRAJ21', 'TRAJ21*01'), ('TRAJ22', 'TRAJ22*01'), ('TRAJ23', 'TRAJ23*01'), ('TRAJ24', 'TRAJ24*01'), ('TRAJ25', 'TRAJ25*01'), ('TRAJ26', 'TRAJ26*01'), ('TRAJ27', 'TRAJ27*01'), ('TRAJ28', 'TRAJ28*01'), ('TRAJ29', 'TRAJ29*01'), ('TRAJ3', 'TRAJ3*01'), ('TRAJ30', 'TRAJ30*01'), ('TRAJ31', 'TRAJ31*01'), ('TRAJ32', 'TRAJ32*01'), ('TRAJ33', 'TRAJ33*01'), ('TRAJ34', 'TRAJ34*01'), ('TRAJ35', 'TRAJ35*01'), ('TRAJ36', 'TRAJ36*01'), ('TRAJ37', 'TRAJ37*01'), ('TRAJ38', 'TRAJ38*01'), ('TRAJ39', 'TRAJ39*01'), ('TRAJ4', 'TRAJ4*01'), ('TRAJ40', 'TRAJ40*01'), ('TRAJ41', 'TRAJ41*01'), ('TRAJ42', 'TRAJ42*01'), ('TRAJ43', 'TRAJ43*01'), ('TRAJ44', 'TRAJ44*01'), ('TRAJ45', 'TRAJ45*01'), ('TRAJ46', 'TRAJ46*01'), ('TRAJ47', 'TRAJ47*01'), ('TRAJ48', 'TRAJ48*01'), ('TRAJ49', 'TRAJ49*01'), ('TRAJ5', 'TRAJ5*01'), ('TRAJ50', 'TRAJ50*01'), ('TRAJ51', 'TRAJ51*01'), ('TRAJ52', 'TRAJ52*01'), ('TRAJ53', 'TRAJ53*01'), ('TRAJ54', 'TRAJ54*01'), ('TRAJ55', 'TRAJ55*01'), ('TRAJ56', 'TRAJ56*01'), ('TRAJ57', 'TRAJ57*01'), ('TRAJ58', 'TRAJ58*01'), ('TRAJ59', 'TRAJ59*01'), ('TRAJ6', 'TRAJ6*01'), ('TRAJ60', 'TRAJ60*01'), ('TRAJ61', 'TRAJ61*01'), ('TRAJ7', 'TRAJ7*01'), ('TRAJ8', 'TRAJ8*01'), ('TRAJ9', 'TRAJ9*01'), ('TRAV1-1', 'TRAV1-1*01'), ('TRAV1-2', 'TRAV1-2*01'), ('TRAV10', 'TRAV10*01'), ('TRAV11', 'TRAV11*01'), ('TRAV12-1', 'TRAV12-1*01'), ('TRAV12-2', 'TRAV12-2*01'), ('TRAV12-3', 'TRAV12-3*01'), ('TRAV13-1', 'TRAV13-1*01'), ('TRAV13-2', 'TRAV13-2*01'), ('TRAV14/DV4', 'TRAV14/DV4*01'), ('TRAV16', 'TRAV16*01'), ('TRAV17', 'TRAV17*01'), ('TRAV18', 'TRAV18*01'), ('TRAV19', 'TRAV19*01'), ('TRAV2', 'TRAV2*01'), ('TRAV20', 'TRAV20*01'), ('TRAV21', 'TRAV21*01'), ('TRAV22', 'TRAV22*01'), ('TRAV23/DV6', 'TRAV23/DV6*01'), ('TRAV24', 'TRAV24*01'), ('TRAV25', 'TRAV25*01'), ('TRAV26-1', 'TRAV26-1*01'), ('TRAV26-2', 'TRAV26-2*01'), ('TRAV27', 'TRAV27*01'), ('TRAV29/DV5', 'TRAV29/DV5*01'), ('TRAV3', 'TRAV3*01'), ('TRAV30', 'TRAV30*01'), ('TRAV34', 'TRAV34*01'), ('TRAV35', 'TRAV35*01'), ('TRAV36/DV7', 'TRAV36/DV7*01'), ('TRAV38-1', 'TRAV38-1*01'), ('TRAV38-2/DV8', 'TRAV38-2/DV8*01'), ('TRAV39', 'TRAV39*01'), ('TRAV4', 'TRAV4*01'), ('TRAV40', 'TRAV40*01'), ('TRAV41', 'TRAV41*01'), ('TRAV5', 'TRAV5*01'), ('TRAV6', 'TRAV6*01'), ('TRAV7', 'TRAV7*01'), ('TRAV8-1', 'TRAV8-1*01'), ('TRAV8-2', 'TRAV8-2*01'), ('TRAV8-3', 'TRAV8-3*01'), ('TRAV8-4', 'TRAV8-4*01'), ('TRAV8-6', 'TRAV8-6*01'), ('TRAV8-7', 'TRAV8-7*01'), ('TRAV9-1', 'TRAV9-1*01'), ('TRAV9-2', 'TRAV9-2*01'), ('TRBD1', 'TRBD1*01'), ('TRBD2', 'TRBD2*01'), ('TRBJ1-1', 'TRBJ1-1*01'), ('TRBJ1-2', 'TRBJ1-2*01'), ('TRBJ1-3', 'TRBJ1-3*01'), ('TRBJ1-4', 'TRBJ1-4*01'), ('TRBJ1-5', 'TRBJ1-5*01'), ('TRBJ1-6', 'TRBJ1-6*01'), ('TRBJ2-1', 'TRBJ2-1*01'), ('TRBJ2-2', 'TRBJ2-2*01'), ('TRBJ2-2P', 'TRBJ2-2P*01'), ('TRBJ2-3', 'TRBJ2-3*01'), ('TRBJ2-4', 'TRBJ2-4*01'), ('TRBJ2-5', 'TRBJ2-5*01'), ('TRBJ2-6', 'TRBJ2-6*01'), ('TRBJ2-7', 'TRBJ2-7*01'), ('TRBV1', 'TRBV1*01'), ('TRBV10-1', 'TRBV10-1*01'), ('TRBV10-2', 'TRBV10-2*01'), ('TRBV10-3', 'TRBV10-3*01'), ('TRBV11-1', 'TRBV11-1*01'), ('TRBV11-2', 'TRBV11-2*01'), ('TRBV11-3', 'TRBV11-3*01'), ('TRBV12-1', 'TRBV12-1*01'), ('TRBV12-2', 'TRBV12-2*01'), ('TRBV12-3', 'TRBV12-3*01'), ('TRBV12-4', 'TRBV12-4*01'), ('TRBV12-5', 'TRBV12-5*01'), ('TRBV13', 'TRBV13*01'), ('TRBV14', 'TRBV14*01'), ('TRBV15', 'TRBV15*01'), ('TRBV16', 'TRBV16*01'), ('TRBV17', 'TRBV17*01'), ('TRBV18', 'TRBV18*01'), ('TRBV19', 'TRBV19*01'), ('TRBV2', 'TRBV2*01'), ('TRBV20-1', 'TRBV20-1*01'), ('TRBV20/OR9-2', 'TRBV20/OR9-2*01'), ('TRBV21-1', 'TRBV21-1*01'), ('TRBV21/OR9-2', 'TRBV21/OR9-2*01'), ('TRBV23-1', 'TRBV23-1*01'), ('TRBV23/OR9-2', 'TRBV23/OR9-2*01'), ('TRBV24-1', 'TRBV24-1*01'), ('TRBV24/OR9-2', 'TRBV24/OR9-2*01'), ('TRBV25-1', 'TRBV25-1*01'), ('TRBV26', 'TRBV26*01'), ('TRBV26/OR9-2', 'TRBV26/OR9-2*01'), ('TRBV27', 'TRBV27*01'), ('TRBV28', 'TRBV28*01'), ('TRBV29-1', 'TRBV29-1*01'), ('TRBV29/OR9-2', 'TRBV29/OR9-2*01'), ('TRBV3-1', 'TRBV3-1*01'), ('TRBV3-2', 'TRBV3-2*01'), ('TRBV30', 'TRBV30*01'), ('TRBV4-1', 'TRBV4-1*01'), ('TRBV4-2', 'TRBV4-2*01'), ('TRBV4-3', 'TRBV4-3*01'), ('TRBV5-1', 'TRBV5-1*01'), ('TRBV5-3', 'TRBV5-3*01'), ('TRBV5-4', 'TRBV5-4*01'), ('TRBV5-5', 'TRBV5-5*01'), ('TRBV5-6', 'TRBV5-6*01'), ('TRBV5-7', 'TRBV5-7*01'), ('TRBV5-8', 'TRBV5-8*01'), ('TRBV6-1', 'TRBV6-1*01'), ('TRBV6-2', 'TRBV6-2*01'), ('TRBV6-3', 'TRBV6-3*01'), ('TRBV6-4', 'TRBV6-4*01'), ('TRBV6-5', 'TRBV6-5*01'), ('TRBV6-6', 'TRBV6-6*01'), ('TRBV6-7', 'TRBV6-7*01'), ('TRBV6-8', 'TRBV6-8*01'), ('TRBV6-9', 'TRBV6-9*01'), ('TRBV7-1', 'TRBV7-1*01'), ('TRBV7-2', 'TRBV7-2*01'), ('TRBV7-3', 'TRBV7-3*01'), ('TRBV7-4', 'TRBV7-4*01'), ('TRBV7-6', 'TRBV7-6*01'), ('TRBV7-7', 'TRBV7-7*01'), ('TRBV7-8', 'TRBV7-8*01'), ('TRBV7-9', 'TRBV7-9*01'), ('TRBV9', 'TRBV9*01'), ('TRDD1', 'TRDD1*01'), ('TRDD2', 'TRDD2*01'), ('TRDD3', 'TRDD3*01'), ('TRDJ1', 'TRDJ1*01'), ('TRDJ2', 'TRDJ2*01'), ('TRDJ3', 'TRDJ3*01'), ('TRDJ4', 'TRDJ4*01'), ('TRDV1', 'TRDV1*01'), ('TRDV2', 'TRDV2*01'), ('TRDV3', 'TRDV3*01'), ('TRGJ1', 'TRGJ1*01'), ('TRGJ2', 'TRGJ2*01'), ('TRGJP', 'TRGJP*01'), ('TRGJP1', 'TRGJP1*01'), ('TRGJP2', 'TRGJP2*01'), ('TRGV1', 'TRGV1*01'), ('TRGV10', 'TRGV10*01'), ('TRGV11', 'TRGV11*01'), ('TRGV2', 'TRGV2*01'), ('TRGV3', 'TRGV3*01'), ('TRGV4', 'TRGV4*01'), ('TRGV5', 'TRGV5*01'), ('TRGV5P', 'TRGV5P*01'), ('TRGV8', 'TRGV8*01'), ('TRGV9', 'TRGV9*01'), ('TRGVA', 'TRGVA*01')])}

    alleles = []
    for g in genes:
        try:
            alleles.append(d[organism][g])
        except KeyError:
            alleles.append(None)
            warnings.warn("{} not found in mapping DB, no allele could be mapped for this gene".format(g), stacklevel=2)
    return(alleles)

def populate_legacy_fields(df, chains =['alpha', 'beta']):
    """
    For instances when we only have v_x_gene and v_j_gene and we need to supply
    'va_countreps' and 'va_gene' for use of tcrdist.subset.TCRsubset()
    """
    if 'alpha' in chains:
        df['va_countreps'] = df['v_a_gene'].copy()
        df['ja_countreps'] = df['j_a_gene'].copy()
        df['va_gene'] = df['v_a_gene'].copy()
        df['ja_gene'] = df['j_a_gene'].copy()
    if 'beta' in chains:
        df['vb_countreps'] = df['v_b_gene'].copy()
        df['jb_countreps'] = df['j_b_gene'].copy()
        df['vb_gene'] = df['v_b_gene'].copy()
        df['jb_gene'] = df['j_b_gene'].copy()
    if 'gamma' in chains:
        df['va_countreps'] = df['v_g_gene'].copy()
        df['ja_countreps'] = df['j_g_gene'].copy()
        df['va_gene'] = df['v_g_gene'].copy()
        df['ja_gene'] = df['j_g_gene'].copy()
    if 'delta' in chains:
        df['vb_countreps'] = df['v_d_gene'].copy()
        df['jb_countreps'] = df['j_d_gene'].copy()
        df['vb_gene'] = df['v_d_gene'].copy()
        df['jb_gene'] = df['j_d_gene'].copy()
    return(df)


import numpy as np
import pandas as pd
import itertools
#from scipy import stats
#import statsmodels.api as sm
#from sklearn.metrics import adjusted_mutual_info_score

from .html_colors import get_html_colors

__all__ = ['plot_pairings']

greek_alpha = '&#x3b1;'
greek_beta  = '&#x3b2;'
greek_gamma = '&#x3b3;'
greek_delta = '&#x3b4;'

segtype2greek_label = { 'VA':'V'+greek_alpha, 'JA':'J'+greek_alpha,
                        'VB':'V'+greek_beta , 'JB':'J'+greek_beta,
                        'VG':'V'+greek_gamma, 'JG':'J'+greek_gamma,
                        'VD':'V'+greek_delta , 'JD':'J'+greek_delta,
                        'v_a_gene':'V'+greek_alpha, 'j_a_gene':'J'+greek_alpha,
                        'v_b_gene':'V'+greek_beta , 'j_b_gene':'J'+greek_beta,
                      'v_g_gene':'V'+greek_gamma, 'j_g_gene':'J'+greek_gamma,
                        'v_d_gene':'V'+greek_delta , 'j_d_gene':'J'+greek_delta}

def _create_svg(cmds, width, height, background_color=None, use_xlink=False):
    out = ''
    extra = '' if not use_xlink else 'xmlns:xlink="http://www.w3.org/1999/xlink"'
    out += '<svg width="{}" height="{}" xmlns="http://www.w3.org/2000/svg" version="1.1" {} >\n'\
              .format(int(width),int(height),extra)
    if background_color:
        out += _rectangle( (0,0), (width, height), background_color, 'white', 0 )
    out += '\n'.join(cmds) + '\n'
    out += '</svg>\n'
    return out

def _rectangle( upper_left, lower_right, fill, stroke, stroke_width=1, dashed=False ):
    stroke_dasharray_style= ';stroke-dasharray:5,5' if dashed else ''
    return '<rect x="{}" y="{}" height="{}" width="{}" style="fill:{};stroke:{};stroke-width:{}{}" />\n'\
        .format( upper_left[0], upper_left[1], lower_right[1]-upper_left[1], lower_right[0]-upper_left[0],
                 fill, stroke, stroke_width, stroke_dasharray_style )

def _make_text( text, lower_left, fontsize,
               extra_tag = None, font_family = "monospace", color = "black", font_weight = "normal" ):
    assert font_weight in ['normal','bold']
    cmd = '<text x="{:.3f}" y="{:.3f}" font-size="{}" font-weight="{}" font-family="{}" fill="{}" xml:space="preserve">{}</text>\n'\
        .format( lower_left[0], lower_left[1], fontsize, font_weight, font_family, color, text )
    return cmd

def _linear_gradient_cmd(gradient_id_counter, x1, y1, x2, y2, offsets, colors, spread_method="pad" ):
    defs_id_prefix = 'TCRGRAD'
    gradient_id_counter += 1
    id = '{}lingrad{:d}'.format(defs_id_prefix, gradient_id_counter)

    stoplines = ''
    assert len(offsets) == len(colors)

    for offset,color in zip( offsets,colors):
        stoplines += """<stop offset="{:.1f}%"   stop-color="{}" stop-opacity="1"/>""".format( offset, color )

    cmd = """<defs>
                    <linearGradient id="{}"
                    x1="{:.1f}%" y1="{:.1f}%"
                    x2="{:.1f}%" y2="{:.1f}%"
                    spreadMethod="{}">
                                    {}
                    </linearGradient>
            </defs>""".format( id, x1, y1, x2, y2, spread_method, stoplines )

    return gradient_id_counter, id, cmd

def _enrichment_glyph_cmds(center, arrow_length, arrow_width, enrichment, add_rectangle=False, box_color='gold' ):
    cmds = []

    num_heads = int(np.floor(np.abs(np.log(enrichment, 2.) ) ))
    if num_heads<1:
        return cmds

    head_sep = 3. * arrow_width
    head_width = 3. * arrow_width
    head_slant = 1.

    if enrichment>1: ## up arrow, text at the top
        line_p0 = [ center[0], center[1] + arrow_length/2. ]
        line_p1 = [ center[0], center[1] - arrow_length/2. ]
        head_step = 1 * head_sep
    else:
        line_p0 = [ center[0], center[1] - arrow_length/2. ]
        line_p1 = [ center[0], center[1] + arrow_length/2. ]
        head_step = -1 * head_sep

    if add_rectangle:
        ## put a nice rounded rectangle around the outside
        rect_x0 = center[0] - head_width - arrow_width
        rect_y0 = center[1] - arrow_length/2 - arrow_width
        rect_width = 2*(head_width+arrow_width)
        rect_height = arrow_length + 2*arrow_width
        rect_round = 3.0 * arrow_width

        cmds.append( '<rect x="{:.3f}" y="{:.3f}" width="{:.3f}" height="{:.3f}" rx="{:.3f}" ry="{:.3f}" stroke="{}" fill="{}"/>'\
                     .format( rect_x0, rect_y0, rect_width, rect_height, rect_round, rect_round,
                              box_color, box_color ))

    ## make the line
    cmds.append( '<line x1="{:.3f}" y1="{:.3f}" x2="{:.3f}" y2="{:.3f}" stroke="black" stroke-width="{}"/>'\
                 .format( line_p0[0], line_p0[1], line_p1[0], line_p1[1], arrow_width ) )

    ## now make the heads
    for head in range(num_heads):
        for xsign in [1,-1]:
            x1 = line_p0[0]
            x2 = x1 + xsign * head_width
            y1 = line_p1[1] + head * head_step
            y2 = y1 + head_slant * head_step

            cmds.append( '<line x1="{:.3f}" y1="{:.3f}" x2="{:.3f}" y2="{:.3f}" stroke="black" stroke-width="{}"/>'\
                         .format( x1,y1,x2,y2, arrow_width ) )

            if False: ## add shifted white line
                y_shift = arrow_width * head_step / head_sep
                cmds.append( '<line x1="{:.3f}" y1="{:.3f}" x2="{:.3f}" y2="{:.3f}" stroke="white" stroke-width="{}"/>'\
                             .format( x1,y1+y_shift,x2,y2+y_shift, arrow_width ) )
    return cmds

def _computeAssociations(df, cols, count_col='Count'):
    res = []
    col1, col2 = cols
    for val1, val2 in itertools.product(df[col1].unique(), df[col2].unique()):
        OR, pvalue, tab = _testAssociation(df, (col1, val1), (col2, val2), count_col=count_col)
        tot = np.sum(tab)
        res.append({'OR':OR,
                    'pvalue':pvalue,
                    'Col0':col1,
                    'Col1':col2,
                    'Val0':val1,
                    'Val1':val2,
                    '00':tab[0, 0],
                    '01':tab[0, 1],
                    '10':tab[1, 0],
                    '11':tab[1, 1]})
    resDf = pd.DataFrame(res)
    # resDf.loc[:, 'qvalue'] = sm.stats.multipletests(resDf['pvalue'].values, method='fdr_bh')[1]
    # resDf = resDf.sort_values(by='pvalue', ascending=True)
    return resDf

def _testAssociation(df, node1, node2, count_col='Count'):
    """Test if the occurence of nodeA paired with nodeB is more/less common than expected.

    Parameters
    ----------
    nodeX : tuple (column, value)
        Specify the node by its column name and the value.

    Returns
    -------
    OR : float
        Odds-ratio associated with the 2x2 contingency table
    pvalue : float
        P-value associated with the Fisher's exact test that H0: OR = 1"""
    col1, val1 = node1
    col2, val2 = node2

    tmp = df[[col1, col2, count_col]].dropna()
    #print(node1, node2, count_col)
    #print(tmp)

    tab = np.zeros((2, 2))
    tab[0, 0] = (((tmp[col1]!=val1) & (tmp[col2]!=val2)) * tmp[count_col]).sum()
    tab[0, 1] = (((tmp[col1]!=val1) & (tmp[col2]==val2)) * tmp[count_col]).sum()
    tab[1, 0] = (((tmp[col1]==val1) & (tmp[col2]!=val2)) * tmp[count_col]).sum()
    tab[1, 1] = (((tmp[col1]==val1) & (tmp[col2]==val2)) * tmp[count_col]).sum()

    # OR, pvalue = stats.fisher_exact(tab)
    OR, pvalue = np.nan, np.nan
    return OR, pvalue, tab

def plot_pairings(cell_df, cols, count_col=None, use_color_gradients=True, other_frequency_threshold=0.01):
    """Diagram depicts the gene-segment pairing structure of the dataset. The four
    genes (cols) are arrayed left to right. Below each gene-type label (eg "VA")
    is a color-stack showing all the TCR clones and how they break down into the different genes for that gene-type. Each clone
    is devoted a constant vertical height in pixels indicated in the text at the top (N pixels in "Nx y-pixel scale"). The curved
    segments joining neighboring gene-stacks show how the two gene distributions pair up, with the thickness of the segments
    corresponding to the number of clones having those two segments (scaled by the indicated y-pixel scale).

    Column names with format: VA or JB (for Valpha and Jbeta) will have the A, B, G, D replaced with its greek character.

    Parameters
    ----------
    cell_df: pd.DataFrame
        Contains gene segment data, one row per clone, optionally with a counts or frequency column.
    cols : list
        List of columns for displaying frequency and pairings, in order from left to right.
    count_col : str
        Optionally provide a count or frequency column for weighting the rows/clones

    Returns
    -------
    raw_svg : str
        Raw SVG txt that can be written to a file.

    Notes
    -----
    For calling this function

    .. code:: python

      import tcrdist as td
      td.plotting.plot_pairings()

    is equivalent to:

    .. code:: python

      import tcrdist as td
      td.gene_pairing_plot.plot_pairings()


    """
    df = cell_df.copy()

    """Not implemented: enrichment should take into account the whole unbiased repertoire"""
    enrichment_glyphs = False

    if count_col is None:
        df = df.assign(Count=1)
        count_col = 'Count'

    params = dict(min_ami_for_colorscale=0.114, # from shuffling experiments
                    max_ami_for_colorscale=0.5,
                    min_entropy_for_colorscale=0.0,
                    max_entropy_for_colorscale=5.0,
                    min_jsd_for_colorscale=0.02259, ## from background vs background comparisons
                    max_jsd_for_colorscale=0.0,
                    min_gene_frequency_for_labels=0.05,
                    reverse_gradients=False)

    gradient_id_counter = 0

    """Pixel units"""
    left_margin = 50
    right_margin = 50
    top_margin = 50
    bottom_margin = 50
    yspacer = 50
    diagram_height = 600
    svg_height = diagram_height + top_margin + bottom_margin

    flat_band = 50
    middle_band = 400
    slope_weight = 100

    ff = 'sans-serif'

    ypixel_scale = diagram_height / df[count_col].sum()
    svg_width = left_margin + right_margin + (len(cols) - 1) * (flat_band + middle_band) + flat_band

    epitope_fontsize = 60
    midpoint = svg_width / 2
    svg_cmds = []
    '''
    svg_cmds.append( _make_text( '{}'.format(epitope, len(tcrs) ),
                                                  [midpoint-0.5*0.6*epitope_fontsize*len(epitope),
                                                   top_margin+epitope_fontsize-20], epitope_fontsize,
                                                  font_family=ff ) )
    '''

    """Compute marginal frequencies for each value within each column"""
    tot = df[count_col].sum()
    counts = {}
    for c in cols:
        """Assign "Other" to all values below a threshold"""
        tmp = df.groupby(c)[count_col].agg(lambda v: np.sum(v) / tot).sort_values(ascending=False)
        ind = tmp.index[tmp < other_frequency_threshold]
        df.loc[df[c].isin(ind), c] = 'Other ' + c
        counts[c] = df.groupby(c)[count_col].agg(lambda v: np.sum(v)).sort_values(ascending=False)
        #print(c)
        #print(counts[c].sum())

    """For each consecutive pair of columns compute 2 x 2 table of frequencies, OR and Fisher's Exact p-value"""
    res = []
    for i in range(len(cols) - 1):
        res.append(_computeAssociations(df, cols=(cols[i], cols[i+1]), count_col=count_col))
    resDf = pd.concat(res, axis=0, ignore_index=True)
    resDf = resDf.sort_values(by='pvalue').set_index(['Val0', 'Val1'])
    #print(resDf[['00', '01', '10', '11']].sum(axis=1))

    for ii in range(len(cols) - 1):
        top_margin = top_margin
        r0 = cols[ii]
        r1 = cols[ii+1]

        x0 = left_margin + ii*( flat_band + middle_band )

        text = segtype2greek_label.get(r0, r0)
        fontsize = 40.
        xtext = x0 + 0.5*flat_band - 0.5*0.6*fontsize*2
        ytext = top_margin + yspacer - 6
        ## hacking
        ytext -= 6
        if ii==0:
            xtext += 8
        svg_cmds.append( _make_text( text, [ xtext, ytext ], fontsize, font_family=ff ) )
        if ii == (len(cols) - 2): ## add the final column label
            text = segtype2greek_label.get(r1, r1)
            xtext = x0+1.5*flat_band-0.5*0.6*fontsize*2+middle_band
            xtext -= 8
            svg_cmds.append( _make_text( text, [ xtext, ytext ], fontsize, font_family=ff ) )

        r0colors = dict(zip(counts[r0].index, get_html_colors(len(counts[r0]))))
        r1colors = dict(zip(counts[r1].index, get_html_colors(len(counts[r1]))))

        a1pixels = {}
        yleft = yspacer + top_margin
        """Nested loops over the alleles (a0, a1) in the pair of columns, r0, r1"""
        for a0, a0count in counts[r0].items():
            y0_right = yspacer + top_margin
            a0color = r0colors[a0]
            for a1, a1count in counts[r1].items():
                a1color = r1colors[a1]
                vj = (a0, a1)
                a1color = r1colors[a1]

                """Frequency of this allele pairing"""
                band_height = resDf.loc[(a0, a1), '11'] * ypixel_scale
                if band_height > 0:
                    """SVG paths with "stroke-width=0" throw errors for image magick"""
                    """Make a spline"""
                    yright = y0_right + a1pixels.get(a1, 0)

                    #line/spline points
                    j_flat_band = flat_band
                    points = [ (np.floor(x0), yleft + 0.5*band_height ),
                               (x0+flat_band, yleft + 0.5*band_height ),
                               (np.ceil(x0 + flat_band + middle_band), yright + 0.5*band_height ),
                               (np.ceil(x0 + flat_band + middle_band + j_flat_band), yright + 0.5*band_height ) ]


                    path1_cmds = 'M {} {} L {} {} M {} {} C {} {}, {} {}, {} {}'\
                        .format( points[0][0], points[0][1], ## start of v-line
                                 points[1][0], points[1][1], ## end point of v-line
                                 points[1][0], points[1][1],
                                 points[1][0] + slope_weight, points[1][1], ## control for spline start
                                 points[2][0] - slope_weight, points[2][1], ## control for spline end
                                 points[2][0], points[2][1] )

                    if use_color_gradients and a0color != a1color:
                        """NOTE: SVG throws an error if trying to apply a gradient to
                        a perfectly horizontal or vertical line due to bounding box issues"""
                        if np.isclose(points[0][1], points[2][1], atol=1e-3):
                            epsilon = 0.005
                        else:
                            epsilon = 0
                        path1a_cmds = 'M {} {} L {} {}'\
                            .format( points[0][0], points[0][1],  ## start of v-line
                                     points[1][0], points[1][1] ) ## end point of v-line
                        svg_cmds.append( '<path d="{}" stroke="{}" stroke-width="{}" fill="none"/>'\
                                                 .format(path1a_cmds, a0color, band_height ) )
                        ## define the gradient
                        path1b_cmds = 'M {} {} C {} {}, {} {}, {} {}'\
                            .format( points[1][0], points[1][1],
                                     points[1][0] + slope_weight, points[1][1], ## control for spline start
                                     points[2][0] - slope_weight, points[2][1], ## control for spline end
                                     points[2][0], points[2][1] + epsilon )
                        #v_line_rhs_fraction = float(flat_band) / (flat_band + middle_band )
                        offsets = [0, 25.0, 75.0, 100]
                        #offsets = [0, 45.0, 55.0, 100]
                        #offsets = [0, 90.0, 99.0, 100]
                        if params['reverse_gradients']:
                            colors = [a1color, a1color, a0color, a0color]
                        else:
                            colors = [a0color, a0color, a1color, a1color]
                        gradient_id_counter, gradient_id, gradient_cmd = _linear_gradient_cmd(gradient_id_counter, 0, 0, 100, 0, offsets, colors )
                        svg_cmds.append( gradient_cmd )
                        svg_cmds.append( '<path d="{}" stroke="url(#{})" stroke-width="{}" fill="none"/>'\
                                                 .format(path1b_cmds, gradient_id, band_height ) )
                    else:
                        svg_cmds.append( '<path d="{}" stroke="{}" stroke-width="{}" fill="none"/>'\
                                                 .format(path1_cmds,a0color, band_height ) )

                    if ii == (len(cols) - 2): ## add the right-most flat band
                        path2_cmds = 'M {} {} L {} {}'\
                            .format( points[2][0], points[2][1], ## start of j-line
                                     points[3][0], points[3][1] ) ## end of j-line

                        svg_cmds.append( '<path d="{}" stroke="{}" stroke-width="{}" fill="none"/>'\
                                                 .format(path2_cmds, a1color, band_height) )

                yleft += band_height
                a1pixels[a1] = a1pixels.get(a1, 0) + band_height
                y0_right += a1count * ypixel_scale

    """Label the alleles in the left stack (and right stack if ii==2)"""
    fontsize = 20
    fontheight = 0.75 * fontsize
    fontwidth =  0.66 * fontsize

    min_height_for_labels = fontheight + 1
    for ii in range(len(cols) - 1):
        x0 = left_margin + ii*( flat_band + middle_band )
        r0 = cols[ii]
        r1 = cols[ii+1]
        r0colors = dict(zip(counts[r0].index, get_html_colors(len(counts[r0]))))
        r1colors = dict(zip(counts[r1].index, get_html_colors(len(counts[r1]))))
        for jj, (r, rcolors) in enumerate([(r0, r0colors), (r1, r1colors)]):
            if jj == 0 and ii > 0:
                continue

            x = x0 + jj*(flat_band + middle_band)
            ystart = yspacer + top_margin
            for a, acount in counts[r].items():
                if acount*ypixel_scale < min_height_for_labels:
                    #print(acount * ypixel_scale)
                    break

                midpoint =  ystart + acount * ypixel_scale * 0.5
                text = a
                lower_left = [x + 2, midpoint + fontheight / 2]

                """Label in white?"""
                bgcolor = rcolors[a]
                textcolor = 'black' if bgcolor!= 'black' else 'white'

                text_width = fontwidth * len(text)
                lower_left_ha = {'left'  : lower_left,
                                 'right' : [ x+flat_band-text_width, midpoint + fontheight/2 ],
                                 'center': [ x+0.5*flat_band-0.5*text_width, midpoint+fontheight/2 ]}
                if jj==0 and ii==0:
                    """Left-most column"""
                    ha = 'left'
                elif jj==1 and ii==(len(cols) - 2):
                    """Right-most column"""
                    ha = 'right'
                else:
                    ha = 'center'
                # print(text, ii, jj, ha, lower_left_ha[ha])
                svg_cmds.append( _make_text( text, lower_left_ha[ha], fontsize, color=textcolor,
                                                              font_family=ff))
                if enrichment_glyphs:
                    """Add an enrichment glyph"""
                    enrich = resDf['OR']
                    if enrich >=2. or enrich <= 0.5:
                        """Add a glyph"""
                        arrow_length = 1.35 * min_height_for_labels
                        arrow_width = 3.5
                        eg_sep = 14.0
                        if 'A' in r:
                            center = [ lower_left_ha[ha][0] + text_width + eg_sep, midpoint ]
                        else:
                            #print rep
                            assert 'B' in r
                            center = [ lower_left_ha[ha][0] - eg_sep, midpoint ]

                        svg_cmds += _enrichment_glyph_cmds(center, arrow_length, arrow_width, enrich)
                ystart += acount * ypixel_scale

    bg_color = None # 'white'
    rawsvg = _create_svg(svg_cmds,
                         width=svg_width,
                         height=svg_height,
                         background_color=bg_color)
    return rawsvg

if __name__ == '__main__':
    np.random.seed(110820)
    n = 50
    df = pd.DataFrame({'VA':np.random.choice(['TRAV14', 'TRAV12', 'TRAV3', 'TRAV23', 'TRAV11', 'TRAV6'], n),
                       'JA':np.random.choice(['TRAJ4', 'TRAJ2', 'TRAJ3','TRAJ5', 'TRAJ21', 'TRAJ13'], n),
                       'VB':np.random.choice(['TRBV14', 'TRBV12', 'TRBV3', 'TRBV23', 'TRBV11', 'TRBV6'], n),
                       'JB':np.random.choice(['TRBJ4', 'TRBJ2', 'TRBJ3','TRBJ5', 'TRBJ21', 'TRBJ13'], n)})
    df = df.assign(Count=1)
    df.loc[:10, 'Count'] = 10
    svg = plot_pairings(df, ['JA', 'VA', 'VB'], count_col='Count')

    """
    import subprocess
    with open('/home/agartlan/gitrepo/test.svg', 'w') as fh:
        fh.write(svg)
    subprocess.check_call('convert -density 200 ../test.svg ../test.png', shell=True)
    """



import pytest



def test_introduction_6():
    """
    Basic Specificity Neighborhoods based on a Hierarchical Clustering
    """
    import pandas as pd
    from tcrdist.repertoire import TCRrep

    df = pd.read_csv("dash.csv")
    tr = TCRrep(cell_df = df, 
                organism = 'mouse', 
                chains = ['beta','alpha'], 
                db_file = 'alphabeta_gammadelta_db.tsv')
    
    from tcrdist.rep_diff import hcluster_diff, member_summ
    from hierdiff import plot_hclust_props

    # diff testing is pasted on binary comparison, so all epitope not 'PA' are set to 'X'
    tr.clone_df['PA'] = ['PA' if x == 'PA' else 'X' for x in tr.clone_df.epitope]
    
    res, Z= hcluster_diff(tr.clone_df, tr.pw_beta, x_cols = ['PA'], count_col = 'count')
    
    res_summary = member_summ(res_df = res, clone_df = tr.clone_df, addl_cols=['epitope'])
    
    res_detailed = pd.concat([res, res_summary], axis = 1)
    
    html = plot_hclust_props(Z,
                title='PA Epitope Example',
                res=res_detailed,
                tooltip_cols=['cdr3_b_aa','v_b_gene', 'j_b_gene','epitope'],
                alpha=0.00001, colors = ['blue','gray'],
                alpha_col='pvalue')

    with open('hierdiff_example.html', 'w') as fh:
        fh.write(html)


import pytest 


def test_fuzzy_div():
    import numpy as np
    from scipy.spatial.distance import squareform
    from tcrdist.diversity import fuzzy_diversity

    n = 100
    counts = np.random.randint(1, 20, size=n)
    dvec = np.round(np.random.rand((n**2 - n) // 2))
    pwmat = squareform(dvec)

    two = fuzzy_diversity(counts, pwmat, order=2, threshold=1)
    two_s = fuzzy_diversity(counts, pwmat, order=2, threshold=1, nsamples=10000, force_sampling=True)
    three = fuzzy_diversity(counts, pwmat, order=3, threshold=1, nsamples=10000)
    four = fuzzy_diversity(counts, pwmat, order=4, threshold=1, nsamples=10000)

    print(two, two_s, three, four)


def test_diversity_example_1():
    import numpy as np
    import pandas as pd
    from tcrdist.repertoire import TCRrep
    from tcrdist.diversity import generalized_simpsons_entropy, simpsons_difference, fuzzy_diversity

    df = pd.read_csv("dash.csv")

    """Compute the Simpson's diversity index (SDI) with 95% CI
    and the "effective number" of TCRs for a TCR repertoire."""
    pb1_div = generalized_simpsons_entropy(df.loc[df['epitope'] == 'PB1', 'count'])

    """The most common SDI is order=2. But see Grabchak et al.
    who generalized the index/entropy for all integer orders and show it is helpful
    to look at a range of orders:

    Grabchak M, Marcon E, Lang G, Zhang Z (2017) The generalized Simpson’s entropy
        is a measure of biodiversity. PLoS ONE 12(3): e0173305.
        https://doi.org/10.1371/journal.pone.0173305"""
    pb1_div = generalized_simpsons_entropy(df.loc[df['epitope'] == 'PB1', 'count'],
                                           orders=np.arange(2, 20),
                                           alpha=0.05)

    """Compare the SDI for two repertoires"""
    pa_div = generalized_simpsons_entropy(df.loc[df['epitope'] == 'PA', 'count'],
                                           orders=np.arange(2, 20),
                                           alpha=0.05)

    div_diff = simpsons_difference(df.loc[df['epitope'] == 'PB1', 'count'],
                                    df.loc[df['epitope'] == 'PA', 'count'],
                                    orders=np.arange(2, 20))

    """Compute a "fuzzy" Simpson's diversity index using the pairwise distance matrix."""
    tr = TCRrep(cell_df = df.loc[df['epitope'] == 'PB1'], 
                organism = 'mouse', 
                chains = ['alpha','beta'], 
                db_file = 'alphabeta_gammadelta_db.tsv')

    fdiv = fuzzy_diversity(tr.clone_df['count'],
                           tr.pw_alpha + tr.pw_beta,
                           order=2,
                           threshold=75)






def test_example_with_report():
    """ 
    Example of TCR radii defined for each TCR in an 
    antigen enriched repertoire, and logo-motif report.
    """
    import os
    import numpy as np
    import pandas as pd
    from tcrdist.repertoire import TCRrep
    from tcrdist.sample import _default_sampler
    from tcrdist.background import get_stratified_gene_usage_frequency
    from tcrdist.centers import calc_radii
    from tcrdist.public import _neighbors_sparse_variable_radius, _neighbors_variable_radius
    from tcrdist.public import TCRpublic
    from tcrdist.ecdf import _plot_manuscript_ecdfs
    import matplotlib.pyplot as plt
        # ANTIGEN ENRICHED REPERTOIRE
        # Load all TCRs tetramer-sorted for the epitope influenza PA epitope
    df = pd.read_csv("dash.csv").query('epitope == "PA"').\
        reset_index(drop = True)
    # Load <df> into a TCRrep instance, to infer CDR1, CDR2, and CDR2.5 region of each clone
    tr = TCRrep(cell_df = df.copy(), 
                organism = 'mouse', 
                chains = ['beta'], 
                db_file = 'alphabeta_gammadelta_db.tsv',
                compute_distances = True)
        # UN-ENRICHED REPERTOIRE
        # For illustration we pull a default sampler for mouse beta chains. 
        # This is used to estimate the gene usage 
        # probabilities P(TRBV = V, TRBJ = J)
    ts = _default_sampler(organism = "mouse", chain = "beta")()
    ts = get_stratified_gene_usage_frequency(ts = ts, replace = True) 
        # Then we synthesize a background using Olga (Sethna et al. 2019), 
        # using the P(TRBV = V, TRBJ = J) for inverse probability weighting.
    df_vj_background = tr.synthesize_vj_matched_background(ts = ts, chain = 'beta')
        # Load <df_vj_background> into a TCRrep instance, to infer CDR1,CDR2,CDR2.5
    trb = TCRrep(cell_df = df_vj_background.copy(), 
                organism = 'mouse', 
                chains = ['beta'], 
                db_file = 'alphabeta_gammadelta_db.tsv',
                compute_distances = False)
        # Take advantage of multiple CPUs
    tr.cpus = 4
        # Compute radii for each TCR that controls neighbor-discovery in the background at 
        # estimate of 1/10^5 inverse probability weighted TCRs.
        # Note we are set <use_sparse> to True, which allows us to take advantage of 
        # multiple cpus and only store distance less than or equal to <max_radius>
    radii, thresholds, ecdfs = \
        calc_radii(tr = tr, 
            tr_bkgd = trb, 
            chain = 'beta', 
            ctrl_bkgd = 10**-5, 
            use_sparse = True, 
            max_radius=50)
        #  Optional, set a maximum radius
    tr.clone_df['radius'] = radii
    tr.clone_df['radius'][tr.clone_df['radius'] > 26] = 26
        # Tabulate index of neighboring clones in the ANTIGEN ENRICHED REPERTOIRE,
        # at each TCR-specific radius   
    tr.clone_df['neighbors'] = _neighbors_variable_radius(
        pwmat = tr.pw_beta, 
        radius_list = tr.clone_df['radius'])
        # Tabulate neighboring sequences in background
    tr.clone_df['background_neighbors'] = _neighbors_sparse_variable_radius(
        csrmat = tr.rw_beta, 
        radius_list = tr.clone_df['radius'])
        # Tabulate number of unique subjects
    tr.clone_df['nsubject']             = tr.clone_df['neighbors'].\
            apply(lambda x: tr.clone_df['subject'].iloc[x].nunique())
        # Score Quasi(Publicity) : True (Quasi-Public), False (private)
    tr.clone_df['qpublic']              = tr.clone_df['nsubject'].\
            apply(lambda x: x > 1)
        # OPTIONAL: HTML Report 
        # Note: you can call TCRpublic() with fixed radius or directly 
        # after tr.clone_df['radius'] is defined. 
    tp = TCRpublic(
        tcrrep = tr, 
        output_html_name = "quasi_public_clones.html")
    tp.fixed_radius = False
        # Generates the HTML report
    rp = tp.report()
        # OPTIONAL: ECDF Figure, against reference
    f1 = _plot_manuscript_ecdfs(
        thresholds = thresholds, 
        ecdf_mat = ecdfs, 
        ylab= 'Proportion of Background TCRs', 
        cdr3_len=tr.clone_df.cdr3_b_aa.str.len(), 
        min_freq=1E-10)
    f1.savefig(os.path.join("", "PA1.png"))
    from tcrdist.ecdf import distance_ecdf
    tresholds, antigen_enriched_ecdf = distance_ecdf(   pwrect = tr.pw_beta,
        thresholds= thresholds,
        weights=None,
        pseudo_count=0,
        skip_diag=False,
        absolute_weight=True)
    # It is straightforward to make a ECDF between antigen enriched TCRs as well:
    antigen_enriched_ecdf[antigen_enriched_ecdf == antigen_enriched_ecdf.min()] = 1E-10
    f2 = _plot_manuscript_ecdfs(
        thresholds = thresholds, 
        ecdf_mat = antigen_enriched_ecdf, 
        ylab= 'Proportion of Antigen Enriched PA TCRs', 
        cdr3_len=tr.clone_df.cdr3_b_aa.str.len(), 
        min_freq=1E-10)
    f2.savefig(os.path.join("", "PA2.png"))


import pandas as pd
import pwseqdist as pw
import numpy as np
from tcrdist.rep_funcs import _pws, _pw

# SEE NEW INTEGRATOIN TEST:  test_integration_of_compute_pw_sparse_out_of_memory
from scipy import sparse
from tcrdist.repertoire import TCRrep
from tcrdist.rep_funcs import  compute_pw_sparse_out_of_memory

def test_pw_rectangular():
    df = pd.read_csv("dash.csv")
    rmat = _pw(
        metric = pw.metrics.nb_vector_tcrdist,
        seqs1 = df.cdr3_b_aa[0:10],
        seqs2 = df.cdr3_b_aa, 
        ncpus=1, 
        uniqify= True, 
        use_numba = True)
    assert rmat.shape == (10,1924)


def test_pws_rectangular_computation():
    metrics = { "cdr3_a_aa" : pw.metrics.nb_vector_tcrdist,
                "pmhc_a_aa" : pw.metrics.nb_vector_tcrdist,
                "cdr2_a_aa" : pw.metrics.nb_vector_tcrdist,
                "cdr1_a_aa" : pw.metrics.nb_vector_tcrdist,
                "cdr3_b_aa" : pw.metrics.nb_vector_tcrdist,
                "pmhc_b_aa" : pw.metrics.nb_vector_tcrdist,
                "cdr2_b_aa" : pw.metrics.nb_vector_tcrdist,
                "cdr1_b_aa" : pw.metrics.nb_vector_tcrdist}

    weights = { 
                "cdr3_a_aa" : 3,
                "pmhc_a_aa" : 1,
                "cdr2_a_aa" : 1,
                "cdr1_a_aa" : 1,
                "cdr3_b_aa" : 3,
                "pmhc_b_aa" : 1,
                "cdr2_b_aa" : 1,
                "cdr1_b_aa" : 1}

    kargs = {   "cdr3_a_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':3, 'ctrim':2, 'fixed_gappos':False},
                "pmhc_a_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True},
                "cdr2_a_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True},
                "cdr1_a_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True},
                "cdr3_b_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':3, 'ctrim':2, 'fixed_gappos':False},
                "pmhc_b_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True},
                "cdr2_b_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True},
                "cdr1_b_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True}}

    df = pd.read_csv("dash2.csv")
    df = df.head(10).copy()
    df2 = pd.read_csv("dash2.csv")
    r = _pws(df = df, df2 = df2, metrics = metrics, weights= weights, kargs=kargs, cpu = 1, store = False)
    assert r['tcrdist'].shape == (10,1924)

def test_dash_tcrdist_fixed_gappos_False():
    import pandas as pd
    import pwseqdist as pw
    from tcrdist.rep_funcs import _pws
    
    # Define metrics for each region
    metrics = { "cdr3_a_aa" : pw.metrics.nb_vector_tcrdist,
                "pmhc_a_aa" : pw.metrics.nb_vector_tcrdist,
                "cdr2_a_aa" : pw.metrics.nb_vector_tcrdist,
                "cdr1_a_aa" : pw.metrics.nb_vector_tcrdist,
                "cdr3_b_aa" : pw.metrics.nb_vector_tcrdist,
                "pmhc_b_aa" : pw.metrics.nb_vector_tcrdist,
                "cdr2_b_aa" : pw.metrics.nb_vector_tcrdist,
                "cdr1_b_aa" : pw.metrics.nb_vector_tcrdist}

    # Define weights
    weights = { 
                "cdr3_a_aa" : 3,
                "pmhc_a_aa" : 1,
                "cdr2_a_aa" : 1,
                "cdr1_a_aa" : 1,
                "cdr3_b_aa" : 3,
                "pmhc_b_aa" : 1,
                "cdr2_b_aa" : 1,
                "cdr1_b_aa" : 1}

    kargs = {   "cdr3_a_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':3, 'ctrim':2, 'fixed_gappos':False},
                "pmhc_a_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True},
                "cdr2_a_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True},
                "cdr1_a_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True},
                "cdr3_b_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':3, 'ctrim':2, 'fixed_gappos':False},
                "pmhc_b_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True},
                "cdr2_b_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True},
                "cdr1_b_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True}}

    df = pd.read_csv("dash2.csv")
    r = _pws(df = df, metrics = metrics, weights= weights, kargs=kargs, cpu = 1, store = False)
    assert r['tcrdist'].shape[0] == 1924
    assert r['tcrdist'].shape[1] == 1924

def test_dash_tcrdist_fixed_gappos_True():
    import pandas as pd
    import pwseqdist as pw
    from tcrdist.rep_funcs import _pws
    
    # Define metrics for each region
    metrics = { "cdr3_a_aa" : pw.metrics.nb_vector_tcrdist,
                "pmhc_a_aa" : pw.metrics.nb_vector_tcrdist,
                "cdr2_a_aa" : pw.metrics.nb_vector_tcrdist,
                "cdr1_a_aa" : pw.metrics.nb_vector_tcrdist,
                "cdr3_b_aa" : pw.metrics.nb_vector_tcrdist,
                "pmhc_b_aa" : pw.metrics.nb_vector_tcrdist,
                "cdr2_b_aa" : pw.metrics.nb_vector_tcrdist,
                "cdr1_b_aa" : pw.metrics.nb_vector_tcrdist}

    # Define weights
    weights = { 
                "cdr3_a_aa" : 3,
                "pmhc_a_aa" : 1,
                "cdr2_a_aa" : 1,
                "cdr1_a_aa" : 1,
                "cdr3_b_aa" : 3,
                "pmhc_b_aa" : 1,
                "cdr2_b_aa" : 1,
                "cdr1_b_aa" : 1}

    kargs = {   "cdr3_a_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':3, 'ctrim':2, 'fixed_gappos':True},
                "pmhc_a_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True},
                "cdr2_a_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True},
                "cdr1_a_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True},
                "cdr3_b_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':3, 'ctrim':2, 'fixed_gappos':True},
                "pmhc_b_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True},
                "cdr2_b_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True},
                "cdr1_b_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True}}

    df = pd.read_csv("dash2.csv")
    r = _pws(df = df, metrics = metrics, weights= weights, kargs=kargs, cpu = 1, store = False)
    assert r['tcrdist'].shape[0] == 1924
    assert r['tcrdist'].shape[1] == 1924
    

def test_dash_nw_metric_fixed_gappos_False():
    import pandas as pd
    import pwseqdist as pw
    from tcrdist.rep_funcs import _pws
    
    # Define metrics for each region
    metrics = { "cdr3_a_aa" : pw.metrics.nw_metric,
                "pmhc_a_aa" : pw.metrics.nw_metric,
                "cdr2_a_aa" : pw.metrics.nw_metric,
                "cdr1_a_aa" : pw.metrics.nw_metric,
                "cdr3_b_aa" : pw.metrics.nw_metric,
                "pmhc_b_aa" : pw.metrics.nw_metric,
                "cdr2_b_aa" : pw.metrics.nw_metric,
                "cdr1_b_aa" : pw.metrics.nw_metric}

    # Define weights
    weights = { 
                "cdr3_a_aa" : 3,
                "pmhc_a_aa" : 1,
                "cdr2_a_aa" : 1,
                "cdr1_a_aa" : 1,
                "cdr3_b_aa" : 3,
                "pmhc_b_aa" : 1,
                "cdr2_b_aa" : 1,
                "cdr1_b_aa" : 1}

    kargs = {   "cdr3_a_aa" : {'use_numba': False},
                "pmhc_a_aa" : {'use_numba': False},
                "cdr2_a_aa" : {'use_numba': False},
                "cdr1_a_aa" : {'use_numba': False},
                "cdr3_b_aa" : {'use_numba': False},
                "pmhc_b_aa" : {'use_numba': False},
                "cdr2_b_aa" : {'use_numba': False},
                "cdr1_b_aa" : {'use_numba': False}}

    df = pd.read_csv("dash2.csv")
    import multiprocessing
    df = df.head(100)
    r = _pws(df = df, metrics = metrics, weights= weights, kargs=kargs, cpu = 1, store = False)
    assert r['tcrdist'].shape[0] == 100
    assert r['tcrdist'].shape[1] == 100



def test_pw():
    df = pd.read_csv("dash2.csv")
    seqs = df['cdr3_b_aa'].values
    r = _pw(metric = pw.metrics.nb_vector_tcrdist, 
            seqs1= seqs,
            seqs2 = None,
            ncpus=1, 
            uniqify=True, 
            use_numba=True, 
            distance_matrix=pw.matrices.tcr_nb_distance_matrix, 
            dist_weight=3, 
            gap_penalty=4, 
            ntrim=3, 
            ctrim=2, 
            fixed_gappos=True)
            
    assert isinstance(r, np.ndarray)

def test_indirect_nw_hamming_metric():
    seqs = ['CASSLDRGEVFF', # Seq1
            'CASSLDRGEVFF', # Seq2 = Seq1 i.e., D(s1,s2) = 0
            'CASSLDAGEVFF', # Seq3 = Seq1 (subsitition A for R) i.e., D(s1,s3) = 1
            'CASSLRGEVFF']  # Seq4 = Seq1 (Delete D for R) i.e., D(s1,s4) = 1 and D(s3,s4) = 2

    result = _pw(metric = pw.metrics.nw_hamming_metric, 
                  seqs1 = seqs, 
                  seqs2 = None, 
                  ncpus=1, 
                  uniqify= True)
    expectation = np.array([[0, 0, 1, 1],[0, 0, 1, 1],[1, 1, 0, 2],[1, 1, 2, 0]])
    assert np.all(result == expectation )

def test_indirect_nw_metric():

    seqs = ['CASSLDRGEVFF', # Seq1
            'CASSLDRGEVFF', # Seq2 = Seq1 i.e., D(s1,s2) = 0
            'CASSLDAGEVFF', # Seq3 = Seq1 (subsitition A for R) i.e., D(s1,s3) = 1
            'CASSLRGEVFF' ] # Seq4 = Seq1 (Delete D for R) i.e., D(s1,s4) = 1 and D(s3,s4) = 2

    result = _pw(metric  = pw.metrics.nw_metric, 
                  seqs1   = seqs, 
                  seqs2   = None, 
                  ncpus   = 1, 
                  uniqify = True)

    assert isinstance(result, np.ndarray)

def test_direct_nw_hamming_metric():
    """
    This example uses nw_hamming_metric. It aligns two sequences and then calculates number of
    mismatched positions. 
    """
    
    seqs = ['CASSLDRGEVFF', # Seq1
            'CASSLDRGEVFF', # Seq2 = Seq1 i.e., D(s1,s2) = 0
            'CASSLDAGEVFF', # Seq3 = Seq1 (subsitition A for R) i.e., D(s1,s3) = 1
            'CASSLRGEVFF']  # Seq4 = Seq1 (Delete D for R) i.e., D(s1,s4) = 1 and D(s3,s4) = 2
    
    result      = pw.apply_pairwise_rect(metric = pw.metrics.nw_hamming_metric, seqs1 = seqs, uniqify= False, ncpus=1)
    expectation = np.array([[0, 0, 1, 1],[0, 0, 1, 1],[1, 1, 0, 2],[1, 1, 2, 0]])
    assert np.all(result == expectation )


def test_indirect_nw_metric_uniquify_returns_full_dimension():

    seqs = ['CASSLDRGEVFF', # Seq1
            'CASSLDRGEVFF', # Seq2 = Seq1 i.e., D(s1,s2) = 0
            'CASSLDAGEVFF', # Seq3 = Seq1 (subsitition A for R) i.e., D(s1,s3) = 1
            'CASSLRGEVFF' ]  # Seq4 = Seq1 (Delete D for R) i.e., D(s1,s4) = 1 and D(s3,s4) = 2

    result = _pw(metric  = pw.metrics.nw_metric, 
                  seqs1   = seqs, 
                  seqs2   = None, 
                  ncpus   = 1, 
                  uniqify = True)

    assert result.shape[1] == 4
    assert isinstance(result, np.ndarray)



def test_integration_of_compute_pw_sparse_out_of_memory():
    """
    INTEGRATION TEST SHOWING HOW MEMORY FUNCTIONS ARE EXPLOITED BY  
    compute_pw_sparse_out_of_memory()
    """
    df = pd.read_csv("dash.csv")

    tr = TCRrep(cell_df = df,               #(2)
                organism = 'mouse',
                chains = ['beta'],
                db_file = 'alphabeta_gammadelta_db.tsv',
                compute_distances = True,
                store_all_cdr = False)

    S,_ = compute_pw_sparse_out_of_memory(tr, matrix_name = "rw_beta", max_distance = 1000)
    # S is a <1920x1920 sparse matrix of type '<class 'numpy.int16'>'
    M = S.todense()
    M[M==-1] = 0
    assert np.all(M == tr.pw_beta)


"""
tcrdist/regex.py 

Regex tools for defining regex patterns from a list of seququences, 
aligned using computepal_motif.
"""

import numpy as np
import pandas as pd
from palmotif import compute_pal_motif
import re

def _list_to_regex_component(l, max_ambiguity = 3):
	""" 
	list of str to regex 

	Parameters
	----------

	l : list
		list of strings
	max_ambiguity : int
		default is 3, more than max results in '.' rather than a set []
	
	Example
	-------
	>>> _list_to_regex_component(['A'])
	'A'
	>>> _list_to_regex_component(['A','T','C'])
	'[ATC]'
	>>> _list_to_regex_component(['A','T','C','G'])
	'.'
	>>> _list_to_regex_component(['A','-','C'])
	'[AC]?'
	"""
	if len(l) < 1:
		s = '.?'
	elif len(l) < 2:
		s = l[0]
	elif len(l) <= max_ambiguity:
		s = f"[{''.join(l)}]"
	else:
		s = '.'
	
	if s == '-':
		s = s.replace('-', '.?') 
	elif s.find('-') != -1:
		s = s.replace('-', '') 
		s = f"{s}?"
	return s


def _index_to_matrix(ind, clone_df, pwmat = None, col = 'cdr3_b_aa', centroid = None ):
	"""
	Example
	-------


	"""
	dfnode   = clone_df.iloc[ind,].copy()

	seqs = dfnode[col].to_list()
	if centroid is None:
		pwnode   = pwmat[ind,:][:,ind].copy()
		iloc_idx = pwnode.sum(axis = 0).argmin() 
		centroid = dfnode[col].to_list()[iloc_idx]
	
	matrix, stats = compute_pal_motif(seqs = seqs, centroid = centroid)
	return matrix


def _matrix_to_regex(matrix, ntrim = 3, ctrim = 2, max_ambiguity = 3):
	"""
	Example
	-------
	>>> import numpy as np
	>>> import pandas as pd
	>>> arr = np.array([[ 0.        ,  3.32910223,  0.        ,  0.        ,  0.41764668,
	         0.        ,  0.        ,  0.        ],
	       [ 0.        ,  0.        ,  0.        ,  0.        ,  0.        ,
	         0.        ,  3.60160675,  4.5849625 ],
	       [ 0.        ,  0.        ,  0.        ,  0.        ,  0.        ,
	         0.        ,  0.        ,  0.        ],
	       [ 0.        ,  0.        ,  0.        ,  0.        ,  0.        ,
	         4.5849625 ,  0.        ,  0.        ],
	       [ 4.5849625 ,  0.        ,  0.41764668,  0.        ,  0.        ,
	         0.        ,  0.        ,  0.        ],
	       [ 0.        ,  0.        ,  0.        ,  0.        ,  2.66666667,
	         0.        ,  0.        ,  0.        ],
	       [ 0.        ,  0.        ,  0.        ,  0.        ,  0.        ,
	         0.        ,  0.        ,  0.        ],
	       [ 0.        ,  0.        ,  0.        , -0.01922274,  0.        ,
	         0.        ,  0.        ,  0.        ]])
	>>> matrix = pd.DataFrame(arr, index = ['A', 'R', 'N', 'D', 'C', 'Q', 'E', 'G'], columns = ['A', 'R', 'N', 'D', 'C', 'Q', 'E', 'G'])
	>>> matrix_to_regex(matrix)
	'(G[AQ]D)'
	"""
	regex_parts =[matrix.iloc[:,][matrix.iloc[:,i] != 0].index.to_list() for i in range(ntrim ,matrix.shape[1] - ctrim)]
	regex_parts_str = [_list_to_regex_component(x, max_ambiguity = max_ambiguity) for x in regex_parts] 
	regex_pattern = f"({''.join(regex_parts_str)})"
	return regex_pattern


def _index_to_regex_str(ind, 
						clone_df, 
						pwmat, 
						centroid = None,
						col = 'cdr3_b_aa', 
						ntrim = 3, 
						ctrim = 2,  
						max_ambiguity = 3):
	"""
	ind : list 
		iloc row index in clone_df
	clone_df : pd.DataFrae
		DataFrame with cdr3 sequence
	pwmat: np.ndarray
		Matrix with pairwise inofrmation
	col : str
		'cdr3_b_aa', 
	ntrim : int
		default 3, 
	ctrim : int
		default 2,  
	max_ambiguity : int
		default 3, the maximum number of amino acids at a position before a '.' wildcard is applied
	"""
	
	mat = _index_to_matrix(ind = ind, clone_df= clone_df, pwmat = pwmat, col = col, centroid = centroid )
	
	regex_str = _matrix_to_regex(matrix = mat, 
		ntrim = ntrim, 
		ctrim = ctrim, 
		max_ambiguity =  max_ambiguity)

	return(regex_str)


def _multi_regex(regex , bkgd_cdr3):
	"""
	Search a regex pattern in a list of string 
	"""
	result = [re.search(string = s, pattern = regex ) for s in bkgd_cdr3] 
	result = [1 if (x is not None) else None for x in result]
	return result

def _index_to_seqs(ind, clone_df, col):
	dfnode   = clone_df.iloc[ind,].copy()
	seqs = dfnode[col].to_list()
	return seqs 
